# flowapp/services/rule_service.py
"""
Service module for rule operations.

This module provides business logic functions for creating, updating,
and managing flow rules, separating these concerns from HTTP handling.
"""

from datetime import datetime, timedelta
from typing import Dict, List, Tuple, Union

from flask import current_app

from flowapp import db, messages
from flowapp.constants import WITHDRAW, RuleOrigin, RuleTypes, ANNOUNCE
from flowapp.models import (
    get_ipv4_model_if_exists,
    get_ipv6_model_if_exists,
    get_rtbh_model_if_exists,
    Flowspec4,
    Flowspec6,
    RTBH,
    Whitelist,
)
from flowapp.models.rules.whitelist import RuleWhitelistCache
from flowapp.models.utils import check_global_rule_limit, check_rule_limit
from flowapp.output import ROUTE_MODELS, Route, announce_route, log_route, RouteSources, log_withdraw
from flowapp.services.base import announce_rtbh_route
from flowapp.services.whitelist_common import (
    Relation,
    add_rtbh_rule_to_cache,
    create_rtbh_from_whitelist_parts,
    subtract_network,
    whitelist_rtbh_rule,
    check_rule_against_whitelists,
)
from flowapp.services.whitelist_service import create_or_update_whitelist
from flowapp.utils import round_to_ten_minutes, get_state_by_time, quote_to_ent


def reactivate_rule(
    rule_type: RuleTypes,
    rule_id: int,
    expires: datetime,
    comment: str,
    user_id: int,
    org_id: int,
    user_email: str,
    org_name: str,
) -> Tuple[Union[RTBH, Flowspec4, Flowspec6], List[str]]:
    """
    Reactivate a rule by setting a new expiration time.

    Args:
        rule_type: Type of rule (RTBH, IPv4, IPv6)
        rule_id: ID of the rule to reactivate
        expires: New expiration datetime
        comment: Updated comment
        user_id: Current user ID
        org_id: Current organization ID
        user_email: User email for logging
        org_name: Organization name for logging

    Returns:
        Tuple containing (rule_model, messages)
    """
    model_name = {RuleTypes.RTBH: RTBH, RuleTypes.IPv4: Flowspec4, RuleTypes.IPv6: Flowspec6}[rule_type]

    model = db.session.get(model_name, rule_id)
    if not model:
        return None, ["Rule not found"]

    flashes = []

    # Check if rule will be reactivated
    state = get_state_by_time(expires)

    # Check global limit
    if state == 1 and check_global_rule_limit(rule_type.value):
        return model, ["global_limit_reached"]

    # Check org limit
    if state == 1 and check_rule_limit(org_id, rule_type=rule_type.value):
        return model, ["limit_reached"]

    # Set new expiration date
    model.expires = expires
    # Set again the active state, if the rule is not whitelisted RTBH
    if rule_type == RuleTypes.RTBH:
        model = check_rtbh_whitelisted(model, user_id, flashes, f"{user_email} / {org_name}")
    else:
        model.rstate_id = state

    model.comment = comment
    db.session.commit()

    route_model = ROUTE_MODELS[rule_type.value]

    if model.rstate_id == 1:
        flashes.append("Rule successfully updated, state set to active.")
        # Announce route
        command = route_model(model, ANNOUNCE)
        route = Route(
            author=f"{user_email} / {org_name}",
            source=RouteSources.UI,
            command=command,
        )
        announce_route(route)
        # Log changes
        log_route(
            user_id,
            model,
            rule_type,
            f"{user_email} / {org_name}",
        )
    else:
        # Withdraw route
        command = route_model(model, WITHDRAW)
        route = Route(
            author=f"{user_email} / {org_name}",
            source=RouteSources.UI,
            command=command,
        )
        announce_route(route)
        # Log changes
        log_withdraw(
            user_id,
            route.command,
            rule_type,
            model.id,
            f"{user_email} / {org_name}",
        )
        if model.rstate_id == 4:
            flashes.append("Rule successfully updated, state set to whitelisted.")
        else:
            flashes.append("Rule successfully updated, state set to inactive.")

    return model, flashes


def create_or_update_ipv4_rule(
    form_data: Dict, user_id: int, org_id: int, user_email: str, org_name: str
) -> Tuple[Flowspec4, str]:
    """
    Create a new IPv4 rule or update an existing one.

    Args:
        form_data: Validated form data
        user_id: Current user ID
        org_id: Current organization ID
        user_email: User email for logging
        org_name: Organization name for logging

    Returns:
        Tuple containing (rule_model, message)
    """
    # Check for existing model
    model = get_ipv4_model_if_exists(form_data, 1)

    if model:
        model.expires = round_to_ten_minutes(form_data["expires"])
        flash_message = "Existing IPv4 Rule found. Expiration time was updated to new value."
    else:
        # Create new model
        model = Flowspec4(
            source=form_data["source"],
            source_mask=form_data["source_mask"],
            source_port=form_data["source_port"],
            destination=form_data["dest"],
            destination_mask=form_data["dest_mask"],
            destination_port=form_data["dest_port"],
            protocol=form_data["protocol"],
            flags=";".join(form_data["flags"]),
            packet_len=form_data["packet_len"],
            fragment=";".join(form_data["fragment"]),
            expires=round_to_ten_minutes(form_data["expires"]),
            comment=quote_to_ent(form_data["comment"]),
            action_id=form_data["action"],
            user_id=user_id,
            org_id=org_id,
            rstate_id=get_state_by_time(form_data["expires"]),
        )
        db.session.add(model)
        flash_message = "IPv4 Rule saved"

    db.session.commit()

    # Announce route if model is in active state
    if model.rstate_id == 1:
        command = messages.create_ipv4(model, ANNOUNCE)
        route = Route(
            author=f"{user_email} / {org_name}",
            source=RouteSources.UI,
            command=command,
        )
        announce_route(route)

    # Log changes
    log_route(
        user_id,
        model,
        RuleTypes.IPv4,
        f"{user_email} / {org_name}",
    )

    return model, flash_message


def create_or_update_ipv6_rule(
    form_data: Dict, user_id: int, org_id: int, user_email: str, org_name: str
) -> Tuple[Flowspec6, str]:
    """
    Create a new IPv6 rule or update an existing one.

    Args:
        form_data: Validated form data
        user_id: Current user ID
        org_id: Current organization ID
        user_email: User email for logging
        org_name: Organization name for logging

    Returns:
        Tuple containing (rule_model, message)
    """
    # Check for existing model
    model = get_ipv6_model_if_exists(form_data, 1)

    if model:
        model.expires = round_to_ten_minutes(form_data["expires"])
        flash_message = "Existing IPv6 Rule found. Expiration time was updated to new value."
    else:
        # Create new model
        model = Flowspec6(
            source=form_data["source"],
            source_mask=form_data["source_mask"],
            source_port=form_data["source_port"],
            destination=form_data["dest"],
            destination_mask=form_data["dest_mask"],
            destination_port=form_data["dest_port"],
            next_header=form_data["next_header"],
            flags=";".join(form_data["flags"]),
            packet_len=form_data["packet_len"],
            expires=round_to_ten_minutes(form_data["expires"]),
            comment=quote_to_ent(form_data["comment"]),
            action_id=form_data["action"],
            user_id=user_id,
            org_id=org_id,
            rstate_id=get_state_by_time(form_data["expires"]),
        )
        db.session.add(model)
        flash_message = "IPv6 Rule saved"

    db.session.commit()

    # Announce routes
    if model.rstate_id == 1:
        command = messages.create_ipv6(model, ANNOUNCE)
        route = Route(
            author=f"{user_email} / {org_name}",
            source=RouteSources.UI,
            command=command,
        )
        announce_route(route)

    # Log changes
    log_route(
        user_id,
        model,
        RuleTypes.IPv6,
        f"{user_email} / {org_name}",
    )

    return model, flash_message


def create_or_update_rtbh_rule(
    form_data: Dict, user_id: int, org_id: int, user_email: str, org_name: str
) -> Tuple[RTBH, List]:
    """
    Create a new RTBH rule or update an existing one.

    Args:
        form_data: Validated form data
        user_id: Current user ID
        org_id: Current organization ID
        user_email: User email for logging
        org_name: Organization name for logging

    Returns:
        Tuple containing (rule_model, message)
    """
    # Check for existing model
    model = get_rtbh_model_if_exists(form_data)
    flashes = []
    if model:
        model.expires = round_to_ten_minutes(form_data["expires"])
        model.rstate_id = get_state_by_time(form_data["expires"])
        flashes.append("Existing RTBH Rule found. Expiration time was updated to new value.")
    else:
        # Create new model
        model = RTBH(
            ipv4=form_data["ipv4"],
            ipv4_mask=form_data["ipv4_mask"],
            ipv6=form_data["ipv6"],
            ipv6_mask=form_data["ipv6_mask"],
            community_id=form_data["community"],
            expires=round_to_ten_minutes(form_data["expires"]),
            comment=quote_to_ent(form_data["comment"]),
            user_id=user_id,
            org_id=org_id,
            rstate_id=get_state_by_time(form_data["expires"]),
        )
        db.session.add(model)
        flashes.append("RTBH Rule saved")

    db.session.commit()

    # rule author for logging and announcing
    author = f"{user_email} / {org_name}"
    # Check if rule is whitelisted
    model = check_rtbh_whitelisted(model, user_id, flashes, author)

    announce_rtbh_route(model, author=author)
    # Log changes
    log_route(user_id, model, RuleTypes.RTBH, author)

    return model, flashes


def check_rtbh_whitelisted(model: RTBH, user_id: int, flashes: List[str], author: str) -> None:
    # Check if rule is whitelisted
    allowed_communities = current_app.config["ALLOWED_COMMUNITIES"]
    if model.community_id in allowed_communities:
        # get all not expired whitelists
        whitelists = db.session.query(Whitelist).filter(Whitelist.expires > datetime.now()).all()
        wl_cache = map_whitelists_to_strings(whitelists)
        results = check_rule_against_whitelists(str(model), wl_cache.keys())
        # check rule against whitelists
        model = evaluate_rtbh_against_whitelists_check_results(user_id, model, flashes, author, wl_cache, results)
    return model


def evaluate_rtbh_against_whitelists_check_results(
    user_id: int,
    model: RTBH,
    flashes: List[str],
    author: str,
    wl_cache: Dict[str, Whitelist],
    results: List[Tuple[str, str, Relation]],
) -> RTBH:
    """
    Evaluate RTBH rule against whitelist check results.
    Process all results for cases where rule is whitelisted by several whitelists.
    """
    for rule, whitelist_key, relation in results:
        if relation == Relation.EQUAL:
            model = whitelist_rtbh_rule(model, wl_cache[whitelist_key])
            msg = f"RTBH Rule {model.id} {model} is equal to active whitelist {whitelist_key}. Rule is whitelisted."
            flashes.append(msg)
            current_app.logger.info(msg)
        elif relation == Relation.SUBNET:
            parts = subtract_network(target=str(model), whitelist=whitelist_key)
            wl_id = wl_cache[whitelist_key].id
            msg = f"RTBH Rule {model.id} {model} is supernet of active whitelist {whitelist_key}. Rule is whitelisted, {len(parts)} subnet rules created."
            flashes.append(msg)
            current_app.logger.info(msg)
            for network in parts:
                new_rule = create_rtbh_from_whitelist_parts(model, wl_id, whitelist_key, network, author, user_id)
                msg = f"Created RTBH rule {new_rule.id} {new_rule} for {network} parted by whitelist {whitelist_key}"
                flashes.append(msg)
                current_app.logger.info(msg)
            model.rstate_id = 4
            add_rtbh_rule_to_cache(model, wl_id, RuleOrigin.USER)
            db.session.commit()
        elif relation == Relation.SUPERNET:
            model = whitelist_rtbh_rule(model, wl_cache[whitelist_key])
            msg = f"RTBH Rule {model.id} {model} is subnet of active whitelist {whitelist_key}. Rule is whitelisted."
            current_app.logger.info(msg)
            flashes.append(msg)
    return model


def map_whitelists_to_strings(whitelists: List[Whitelist]) -> Dict[str, Whitelist]:
    return {str(w): w for w in whitelists}


def delete_rule(
    rule_type: RuleTypes, rule_id: int, user_id: int, user_email: str, org_name: str, allowed_rule_ids: List[int] = None
) -> Tuple[bool, str]:
    """
    Delete a rule with the given id and type.

    Args:
        rule_type: Type of rule (RTBH, IPv4, IPv6)
        rule_id: ID of the rule to delete
        user_id: Current user ID
        user_email: User email for logging
        org_name: Organization name for logging
        allowed_rule_ids: List of rule IDs the user is allowed to delete, None means no restriction

    Returns:
        Tuple containing (success, message)
    """
    model_class = {RuleTypes.RTBH: RTBH, RuleTypes.IPv4: Flowspec4, RuleTypes.IPv6: Flowspec6}[rule_type]

    route_model = ROUTE_MODELS[rule_type.value]

    model = db.session.get(model_class, rule_id)
    if not model:
        return False, "Rule not found"

    # Check permission if allowed_rule_ids is provided
    if allowed_rule_ids is not None and model.id not in allowed_rule_ids:
        return False, "You cannot delete this rule"

    # Withdraw route
    command = route_model(model, WITHDRAW)
    route = Route(
        author=f"{user_email} / {org_name}",
        source=RouteSources.UI,
        command=command,
    )
    announce_route(route)

    # Log withdrawal
    log_withdraw(
        user_id,
        route.command,
        rule_type,
        model.id,
        f"{user_email} / {org_name}",
    )

    # Special handling for RTBH rules
    if rule_type == RuleTypes.RTBH:
        current_app.logger.debug(f"Deleting RTBH rule {rule_id} from cache")
        RuleWhitelistCache.delete_by_rule_id(rule_id)

    # Delete from database
    db.session.delete(model)
    db.session.commit()

    return True, "Rule deleted successfully"


def delete_rtbh_and_create_whitelist(
    rule_id: int,
    user_id: int,
    org_id: int,
    user_email: str,
    org_name: str,
    allowed_rule_ids: List[int] = None,
    whitelist_expires: datetime = None,
) -> Tuple[bool, List[str], Union[Whitelist, None]]:
    """
    Delete an RTBH rule and create a whitelist entry from it.

    Args:
        rule_id: ID of the RTBH rule to delete
        user_id: Current user ID
        org_id: Current organization ID
        user_email: User email for logging
        org_name: Organization name for logging
        allowed_rule_ids: List of rule IDs the user is allowed to delete
        whitelist_expires: Expiration time for the whitelist entry (default: 7 days from now)

    Returns:
        Tuple containing (success, messages, whitelist_model)
    """
    messages = []

    # First get the RTBH rule to extract its data
    model = db.session.get(RTBH, rule_id)
    if not model:
        return False, ["RTBH rule not found"], None

    # Check permission if allowed_rule_ids is provided
    if allowed_rule_ids is not None and model.id not in allowed_rule_ids:
        return False, ["You cannot delete this rule"], None

    # Extract data for whitelist
    if model.ipv4:
        ip = model.ipv4
        mask = model.ipv4_mask
    elif model.ipv6:
        ip = model.ipv6
        mask = model.ipv6_mask
    else:
        return False, ["RTBH rule has no IP address"], None

    # Set default whitelist expiration time if not provided
    if whitelist_expires is None:
        whitelist_expires = datetime.now() + timedelta(days=7)

    # Prepare whitelist data
    # Create base comment
    comment_text = f"Created from RTBH rule {model} {rule_id}"
    # Append the rule's comment only if it exists
    if model.comment:
        comment_text += f": {model.comment}"

    whitelist_data = {
        "ip": ip,
        "mask": mask,
        "expires": whitelist_expires,
        "comment": comment_text,
    }

    # Delete the RTBH rule
    success, delete_message = delete_rule(
        rule_type=RuleTypes.RTBH,
        rule_id=rule_id,
        user_id=user_id,
        user_email=user_email,
        org_name=org_name,
        allowed_rule_ids=allowed_rule_ids,
    )

    if not success:
        return False, [delete_message], None

    messages.append(delete_message)

    # Create the whitelist entry
    try:
        whitelist_model, whitelist_messages = create_or_update_whitelist(
            form_data=whitelist_data, user_id=user_id, org_id=org_id, user_email=user_email, org_name=org_name
        )
        messages.extend(whitelist_messages)
        return True, messages, whitelist_model
    except Exception as e:
        current_app.logger.exception(f"Error creating whitelist entry: {e}")
        messages.append(f"Rule deleted but failed to create whitelist: {str(e)}")
        return False, messages, None


def delete_expired_rules() -> Dict[str, int]:
    """
    Delete all expired rules older than EXPIRATION_THRESHOLD days.
    Only deletes rules in withdrawn or deleted state.

    Returns:
        Dictionary with deletion counts per rule type
    """
    current_time = datetime.now()
    expiration_threshold = current_app.config.get("EXPIRATION_THRESHOLD", 30)
    deletion_date = current_time - timedelta(days=expiration_threshold)

    deletion_counts = {"rtbh": 0, "ipv4": 0, "ipv6": 0, "total": 0}

    model_map = {
        "rtbh": (RTBH, RuleTypes.RTBH),
        "ipv4": (Flowspec4, RuleTypes.IPv4),
        "ipv6": (Flowspec6, RuleTypes.IPv6),
    }

    for rule_type, (model_class, rule_enum) in model_map.items():
        # Get IDs of rules to delete
        expired_rule_ids = [
            r.id
            for r in db.session.query(model_class.id)
            .filter(
                model_class.expires < deletion_date, model_class.rstate_id.in_([2, 3])  # withdrawn or deleted state
            )
            .all()
        ]

        if not expired_rule_ids:
            current_app.logger.info(f"No expired {model_class.__name__} rules to delete")
            continue

        # Clean up whitelist cache first
        cache_deleted = 0
        for rule_id in expired_rule_ids:
            cache_deleted += RuleWhitelistCache.delete_by_rule_id(rule_id)

        if cache_deleted:
            current_app.logger.info(
                f"Deleted {cache_deleted} cache entries for {len(expired_rule_ids)} {model_class.__name__} rules"
            )

        # Bulk delete the rules
        deleted = (
            db.session.query(model_class).filter(model_class.id.in_(expired_rule_ids)).delete(synchronize_session=False)
        )

        deletion_counts[rule_type] = deleted
        deletion_counts["total"] += deleted

        current_app.logger.info(
            f"Deleted {deleted} expired {model_class.__name__} rules " f"(older than {expiration_threshold} days)"
        )

    # Commit all deletions at once
    if deletion_counts["total"] > 0:
        try:
            db.session.commit()
            current_app.logger.info(
                f"Successfully deleted {deletion_counts['total']} expired rules: "
                f"RTBH={deletion_counts['rtbh']}, "
                f"IPv4={deletion_counts['ipv4']}, "
                f"IPv6={deletion_counts['ipv6']}"
            )
        except Exception as e:
            db.session.rollback()
            current_app.logger.error(f"Error committing rule deletions: {e}")
            raise
    else:
        current_app.logger.info("No expired rules found to delete")

    return deletion_counts



"""
Rule forms for the flowapp application.
"""

from .base import IPForm
from .ipv4 import IPv4Form
from .ipv6 import IPv6Form
from .rtbh import RTBHForm
from .whitelist import WhitelistForm

__all__ = [
    "IPForm",
    "IPv4Form",
    "IPv6Form",
    "RTBHForm",
    "WhitelistForm",
]



"""
IPv4 rule form for the flowapp application.
"""

from wtforms import StringField, IntegerField, SelectField, SelectMultipleField
from wtforms.validators import Optional, NumberRange, DataRequired

from ...constants import IPV4_PROTOCOL, IPV4_FRAGMENT
from ...validators import IPv4Address
from .base import IPForm


class IPv4Form(IPForm):
    """
    IPv4 form object
    """

    def __init__(self, *args, **kwargs):
        super(IPv4Form, self).__init__(*args, **kwargs)
        self.net_ranges = None

    zero_address = "0.0.0.0"
    source = StringField(
        "Source address",
        validators=[Optional(), IPv4Address(message="provide valid IPv4 adress")],
    )

    source_mask = IntegerField(
        "Source mask (bits)",
        validators=[
            Optional(),
            NumberRange(min=0, max=32, message="invalid mask value (0-32)"),
        ],
    )

    dest = StringField(
        "Destination address",
        validators=[Optional(), IPv4Address(message="provide valid IPv4 adress")],
    )

    dest_mask = IntegerField(
        "Destination mask (bits)",
        validators=[
            Optional(),
            NumberRange(min=0, max=32, message="invalid mask value (0-32)"),
        ],
    )

    protocol = SelectField(
        "Protocol",
        choices=[(pr, pr.upper()) for pr in IPV4_PROTOCOL.keys()],
        validators=[DataRequired()],
    )

    fragment = SelectMultipleField(
        "Fragment",
        choices=[(frv, frk.upper()) for frk, frv in IPV4_FRAGMENT.items()],
        validators=[Optional()],
    )

    def validate_ipv_specific(self):
        """
        validate protocol and flags, set error message if validation fails
        :return: boolean validation result
        """

        if self.flags.data and self.protocol.data and len(self.flags.data) > 0 and self.protocol.data != "tcp":
            self.flags.errors.append("Can not set TCP flags for protocol {} !".format(self.protocol.data.upper()))
            return False
        return True



from datetime import datetime
from operator import ge, lt
from flowapp import constants, db, messages
from flowapp.constants import ANNOUNCE, WITHDRAW
from flowapp.models import RTBH, Flowspec4, Flowspec6
from flowapp.output import Route, RouteSources, announce_route


def announce_rtbh_route(model: RTBH, author: str) -> None:
    """
    Announce RTBH route if rule is in active state
    """
    if model.rstate_id == 1:
        command = messages.create_rtbh(model, ANNOUNCE)
        route = Route(
            author=author,
            source=RouteSources.UI,
            command=command,
        )
        announce_route(route)


def withdraw_rtbh_route(model: RTBH) -> None:
    """
    Withdraw RTBH route if rule is in whitelist state
    """
    if model.rstate_id == 4:
        command = messages.create_rtbh(model, WITHDRAW)
        route = Route(
            author=model.get_author(),
            source=RouteSources.UI,
            command=command,
        )
        announce_route(route)


def announce_all_routes(action=constants.ANNOUNCE):
    """
    get routes from db and send it to ExaBGB api

    @TODO take the request away, use some kind of messaging (maybe celery?)
    :param action: action with routes - announce valid routes or withdraw expired routes
    """
    today = datetime.now()
    comp_func = ge if action == constants.ANNOUNCE else lt

    rules4 = (
        db.session.query(Flowspec4)
        .filter(Flowspec4.rstate_id == 1)
        .filter(comp_func(Flowspec4.expires, today))
        .order_by(Flowspec4.expires.desc())
        .all()
    )
    rules6 = (
        db.session.query(Flowspec6)
        .filter(Flowspec6.rstate_id == 1)
        .filter(comp_func(Flowspec6.expires, today))
        .order_by(Flowspec6.expires.desc())
        .all()
    )
    rules_rtbh = (
        db.session.query(RTBH)
        .filter(RTBH.rstate_id == 1)
        .filter(comp_func(RTBH.expires, today))
        .order_by(RTBH.expires.desc())
        .all()
    )

    messages_v4 = [messages.create_ipv4(rule, action) for rule in rules4]
    messages_v6 = [messages.create_ipv6(rule, action) for rule in rules6]
    messages_rtbh = [messages.create_rtbh(rule, action) for rule in rules_rtbh]

    messages_all = []
    messages_all.extend(messages_v4)
    messages_all.extend(messages_v6)
    messages_all.extend(messages_rtbh)

    author_action = "announce all" if action == constants.ANNOUNCE else "withdraw all expired"

    for command in messages_all:
        route = Route(
            author=f"System call / {author_action} rules",
            source=RouteSources.UI,
            command=command,
        )
        announce_route(route)

    if action == constants.WITHDRAW:
        for ruleset in [rules4, rules6, rules_rtbh]:
            for rule in ruleset:
                rule.rstate_id = 2

        db.session.commit()



V_PREFIX = "/api/v1"


def test_token(client, jwt_token):
    """
    test that token authorization works
    """
    req = client.get(f"{V_PREFIX}/test_token", headers={"x-access-token": jwt_token})

    assert req.status_code == 400


def test_withnout_token(client):
    """
    test that without token authorization return 401
    """
    req = client.get(f"{V_PREFIX}/test_token")

    assert req.status_code == 400


def test_rules(client, db, jwt_token):
    """
    test that there is one ipv4 rule created in the first test
    """
    req = client.get(f"{V_PREFIX}/rules", headers={"x-access-token": jwt_token})

    assert req.status_code == 400



from functools import wraps
from typing import List, Optional
from flask import current_app, redirect, request, url_for, session, abort

from flowapp import __version__, db, validators
from flowapp.models import Flowspec4, Flowspec6, RTBH, Whitelist, get_user_nets


def auth_required(f):
    """
    auth required decorator
    """

    @wraps(f)
    def decorated(*args, **kwargs):
        user = get_user()
        session["app_version"] = __version__
        if not user:
            if current_app.config.get("SSO_AUTH"):
                current_app.logger.warning("SSO AUTH SET")
                return redirect("/login")

            if current_app.config.get("HEADER_AUTH", False):
                return redirect("/ext_login")

            if current_app.config.get("LOCAL_AUTH"):
                return redirect(url_for("local_login"))

        return f(*args, **kwargs)

    return decorated


def admin_required(f):
    """
    decorator for admin only endpoints
    """

    @wraps(f)
    def decorated(*args, **kwargs):
        if 3 not in session["user_role_ids"]:
            return redirect(url_for("index"))
        return f(*args, **kwargs)

    return decorated


def user_or_admin_required(f):
    """
    decorator for admin/user endpoints
    """

    @wraps(f)
    def decorated(*args, **kwargs):
        if not all(i > 1 for i in session["user_role_ids"]):
            return redirect(url_for("index"))
        return f(*args, **kwargs)

    return decorated


def localhost_only(f):
    """
    auth required decorator
    """

    @wraps(f)
    def decorated(*args, **kwargs):
        remote = request.remote_addr
        localv4 = current_app.config.get("LOCAL_IP")
        localv6 = current_app.config.get("LOCAL_IP6")
        if remote != localv4 and remote != localv6:
            current_app.logger.warning(f"AUTH LOCAL ONLY FAIL FROM {remote} / local adresses [{localv4}, {localv6}]")
            abort(403)  # Forbidden
        return f(*args, **kwargs)

    return decorated


def check_access_rights(current_user, model_id):
    """
    Check if the current user has right to edit/delete certain model data
    Used in API
    Returns true if the user is owner of the record or if the user is admin
    :param current_user: api current user object
    :param model_id: user_id from the model data
    :return: boolean
    """
    if model_id == current_user["id"]:
        return True

    if max(current_user["role_ids"]) == 3:
        return True

    return False


def check_access_rights_gui(current_user_id, current_user_roles, model_id):
    """
    Check if the current user has right to edit/delete certain model data
    Used in GUI - rules.py etc.
    Returns true if the user is owner of the record or if the user is admin
    :param current_user_id: current user id
    :param current_user_roles: current user roles
    :param model_id: user_id from the model data
    :return: boolean
    """
    if model_id == current_user_id:
        return True

    if max(current_user_roles) == 3:
        return True

    return False


def is_admin(current_user_roles):
    """
    Check if the current user is admin
    :param current_user_roles: current user roles
    :return: boolean
    """
    if max(current_user_roles) == 3:
        return True

    return False


def get_user():
    """
    get user from session or return None
    """
    return session.get("user_uuid", None)


def get_user_allowed_rule_ids(rule_type: str, user_id: int, user_role_ids: List[int]) -> List[int]:
    """
    Get list of rule IDs that the user is allowed to modify.

    For admin users (role_id 3), returns all rule IDs of the given type.
    For regular users, returns only rules within their network ranges.

    Args:
        rule_type: Type of rule ('ipv4', 'ipv6', 'rtbh', 'whitelist')
        user_id: Current user's ID
        user_role_ids: List of user's role IDs

    Returns:
        List of rule IDs the user can modify
    """
    # Admin users can modify any rules
    if 3 in user_role_ids:
        if rule_type == "ipv4":
            return [r.id for r in db.session.query(Flowspec4.id).all()]
        elif rule_type == "ipv6":
            return [r.id for r in db.session.query(Flowspec6.id).all()]
        elif rule_type == "rtbh":
            return [r.id for r in db.session.query(RTBH.id).all()]
        elif rule_type == "whitelist":
            return [r.id for r in db.session.query(Whitelist.id).all()]
        return []

    # Regular users - filter by network ranges
    net_ranges = get_user_nets(user_id)

    if rule_type == "ipv4":
        rules = db.session.query(Flowspec4).all()
        filtered_rules = validators.filter_rules_in_network(net_ranges, rules)
        return [r.id for r in filtered_rules]

    elif rule_type == "ipv6":
        rules = db.session.query(Flowspec6).all()
        filtered_rules = validators.filter_rules_in_network(net_ranges, rules)
        return [r.id for r in filtered_rules]

    elif rule_type == "rtbh":
        rules = db.session.query(RTBH).all()
        filtered_rules = validators.filter_rtbh_rules(net_ranges, rules)
        return [r.id for r in filtered_rules]

    elif rule_type == "whitelist":
        rules = db.session.query(Whitelist).all()
        filtered_rules = validators.filter_rules_in_network(net_ranges, rules)
        return [r.id for r in filtered_rules]

    return []


def check_user_can_modify_rule(
    rule_id: int, rule_type: str, user_id: Optional[int] = None, user_role_ids: Optional[List[int]] = None
) -> bool:
    """
    Check if the current user can modify a specific rule.

    Args:
        rule_id: ID of the rule to check
        rule_type: Type of rule ('ipv4', 'ipv6', 'rtbh', 'whitelist')
        user_id: User ID (defaults to session user_id)
        user_role_ids: User role IDs (defaults to session user_role_ids)

    Returns:
        True if user can modify the rule, False otherwise
    """
    if user_id is None:
        user_id = session.get("user_id")
    if user_role_ids is None:
        user_role_ids = session.get("user_role_ids", [])

    # Admin users can modify any rules
    if 3 in user_role_ids:
        return True

    # Check if rule_id is in allowed rules for this user
    allowed_ids = get_user_allowed_rule_ids(rule_type, user_id, user_role_ids)
    return rule_id in allowed_ids



"""
Base form fields for the flowapp application.
"""

from wtforms import widgets
from wtforms.fields import DateTimeField
from ..utils import parse_api_time


class MultiFormatDateTimeLocalField(DateTimeField):
    """
    Same as :class:`~wtforms.fields.DateTimeField`, but represents an
    ``<input type="datetime-local">``.

    Custom implementation uses default HTML5 format for parsing the field.
    It's possible to use multiple formats - used in API.

    """

    widget = widgets.DateTimeLocalInput()

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("format", "%Y-%m-%dT%H:%M")
        self.unlimited = kwargs.pop("unlimited", False)
        self.pref_format = None
        super().__init__(*args, **kwargs)

    def process_formdata(self, valuelist):
        if not valuelist or (len(valuelist) == 1 and not valuelist[0]):
            return None

        # with unlimited field we do not need to parse the empty value
        if self.unlimited and len(valuelist) == 1 and len(valuelist[0]) == 0:
            self.data = None
            return None

        date_str = " ".join((str(val) for val in valuelist))

        try:
            result, pref_format = parse_api_time(date_str)
        except TypeError:
            raise ValueError(self.gettext("Not a valid datetime value."))

        if result:
            self.data = result
            self.pref_format = pref_format
        else:
            self.data = None
            self.pref_format = None
            raise ValueError(self.gettext("Not a valid datetime value."))



import pytest
from datetime import datetime, timedelta
from werkzeug.datastructures import MultiDict
from flowapp.forms import (
    UserForm,
    BulkUserForm,
    ApiKeyForm,
    MachineApiKeyForm,
    OrganizationForm,
    ActionForm,
    ASPathForm,
    CommunityForm,
    RTBHForm,
    IPv4Form,
    IPv6Form,
    WhitelistForm,
)


def create_form_data(data):
    """Helper function to create proper form data format"""
    processed_data = {}
    for key, value in data.items():
        if isinstance(value, list):
            processed_data[key] = [str(v) for v in value]
        else:
            processed_data[key] = value
    return MultiDict(processed_data)


@pytest.fixture
def valid_datetime():
    return (datetime.now() + timedelta(days=1)).strftime("%Y-%m-%dT%H:%M")


@pytest.fixture
def sample_network_ranges():
    return ["192.168.0.0/16", "2001:db8::/32"]


class TestUserForm:
    @pytest.fixture
    def mock_choices(self):
        return {
            "role_ids": [(1, "Admin"), (2, "User"), (3, "Guest")],
            "org_ids": [(1, "Org1"), (2, "Org2"), (3, "Org3")],
        }

    @pytest.fixture
    def valid_user_data(self):
        return {
            "uuid": "test@example.com",
            "email": "user@example.com",
            "name": "Test User",
            "phone": "123456789",
            "role_ids": ["2"],
            "org_ids": ["1"],
        }

    def test_valid_user_form(self, app, valid_user_data, mock_choices):
        with app.test_request_context():
            form_data = create_form_data(valid_user_data)
            form = UserForm(formdata=form_data)
            form.role_ids.choices = mock_choices["role_ids"]
            form.org_ids.choices = mock_choices["org_ids"]

            if not form.validate():
                print("Validation errors:", form.errors)

            assert form.validate()

    def test_invalid_email(self, app, mock_choices):
        with app.test_request_context():
            form_data = create_form_data({"uuid": "invalid-email", "role_ids": ["2"], "org_ids": ["1"]})
            form = UserForm(formdata=form_data)
            form.role_ids.choices = mock_choices["role_ids"]
            form.org_ids.choices = mock_choices["org_ids"]

            assert not form.validate()
            assert "Please provide valid email" in form.uuid.errors


class TestRTBHForm:
    @pytest.fixture
    def mock_community_choices(self):
        return [(1, "Community1"), (2, "Community2")]

    @pytest.fixture
    def valid_rtbh_data(self, valid_datetime):
        return {
            "ipv4": "192.168.1.0",
            "ipv4_mask": 24,
            "community": "1",
            "expires": valid_datetime,
            "comment": "Test RTBH rule",
        }

    def test_valid_ipv4_rtbh(self, app, valid_rtbh_data, sample_network_ranges, mock_community_choices):
        with app.test_request_context():
            form_data = create_form_data(valid_rtbh_data)
            form = RTBHForm(formdata=form_data)
            form.net_ranges = sample_network_ranges
            form.community.choices = mock_community_choices
            assert form.validate()


class TestIPv4Form:
    @pytest.fixture
    def mock_action_choices(self):
        return [(1, "Accept"), (2, "Drop"), (3, "Reject")]

    @pytest.fixture
    def valid_ipv4_data(self, valid_datetime):
        return {
            "source": "192.168.1.0",
            "source_mask": "24",
            "protocol": "tcp",
            "action": "1",
            "expires": valid_datetime,
        }

    def test_valid_ipv4_rule(self, app, valid_ipv4_data, sample_network_ranges, mock_action_choices):
        with app.test_request_context():
            form_data = create_form_data(valid_ipv4_data)
            form = IPv4Form(formdata=form_data)
            form.net_ranges = sample_network_ranges
            form.action.choices = mock_action_choices

            if not form.validate():
                print("Validation errors:", form.errors)

            assert form.validate()

    def test_invalid_protocol_flags(self, app, valid_datetime, sample_network_ranges, mock_action_choices):
        with app.test_request_context():
            form_data = create_form_data(
                {
                    "source": "192.168.1.0",
                    "source_mask": "24",
                    "protocol": "udp",
                    "flags": ["syn"],
                    "action": "1",
                    "expires": valid_datetime,
                }
            )
            form = IPv4Form(formdata=form_data)
            form.net_ranges = sample_network_ranges
            form.action.choices = mock_action_choices
            assert not form.validate()


class TestIPv6Form:
    @pytest.fixture
    def mock_action_choices(self):
        return [(1, "Accept"), (2, "Drop"), (3, "Reject")]

    @pytest.fixture
    def valid_ipv6_data(self, valid_datetime):
        return {
            "source": "2001:db8::",  # Network aligned address within allowed range
            "source_mask": "32",  # Matching the organization's prefix length
            "next_header": "tcp",
            "action": "1",
            "expires": valid_datetime,
        }

    def test_valid_ipv6_rule(self, app, valid_ipv6_data, sample_network_ranges, mock_action_choices):
        with app.test_request_context():
            form_data = create_form_data(valid_ipv6_data)
            form = IPv6Form(formdata=form_data)
            form.net_ranges = sample_network_ranges
            form.action.choices = mock_action_choices

            if not form.validate():
                print("Validation errors:", form.errors)

            assert form.validate()

    def test_invalid_next_header_flags(self, app, valid_datetime, sample_network_ranges, mock_action_choices):
        with app.test_request_context():
            form_data = create_form_data(
                {
                    "source": "2001:db8::",
                    "source_mask": "32",
                    "next_header": "udp",
                    "flags": ["syn"],
                    "action": "1",
                    "expires": valid_datetime,
                }
            )
            form = IPv6Form(formdata=form_data)
            form.net_ranges = sample_network_ranges
            form.action.choices = mock_action_choices
            assert not form.validate()

    def test_address_outside_range(self, app, valid_datetime, sample_network_ranges, mock_action_choices):
        """Test validation fails when address is outside allowed ranges"""
        with app.test_request_context():
            form_data = create_form_data(
                {
                    "source": "2001:db9::",  # Different prefix
                    "source_mask": "32",
                    "next_header": "tcp",
                    "action": "1",
                    "expires": valid_datetime,
                }
            )
            form = IPv6Form(formdata=form_data)
            form.net_ranges = sample_network_ranges
            form.action.choices = mock_action_choices
            assert not form.validate()
            assert any("must be in organization range" in error for error in form.source.errors)

    def test_destination_address(self, app, valid_datetime, sample_network_ranges, mock_action_choices):
        """Test validation with destination address instead of source"""
        with app.test_request_context():
            form_data = create_form_data(
                {
                    "dest": "2001:db8::",
                    "dest_mask": "32",
                    "next_header": "tcp",
                    "action": "1",
                    "expires": valid_datetime,
                }
            )
            form = IPv6Form(formdata=form_data)
            form.net_ranges = sample_network_ranges
            form.action.choices = mock_action_choices

            if not form.validate():
                print("Validation errors:", form.errors)

            assert form.validate()

    def test_both_source_and_dest(self, app, valid_datetime, sample_network_ranges, mock_action_choices):
        """Test validation with both source and destination addresses"""
        with app.test_request_context():
            form_data = create_form_data(
                {
                    "source": "2001:db8::",
                    "source_mask": "32",
                    "dest": "2001:db8:1::",
                    "dest_mask": "48",
                    "next_header": "tcp",
                    "action": "1",
                    "expires": valid_datetime,
                }
            )
            form = IPv6Form(formdata=form_data)
            form.net_ranges = sample_network_ranges
            form.action.choices = mock_action_choices

            if not form.validate():
                print("Validation errors:", form.errors)

            assert form.validate()

    def test_tcp_flags(self, app, valid_datetime, sample_network_ranges, mock_action_choices):
        """Test validation with TCP flags (should be valid with TCP)"""
        with app.test_request_context():
            form_data = create_form_data(
                {
                    "source": "2001:db8::",
                    "source_mask": "32",
                    "next_header": "tcp",
                    "flags": ["SYN", "ACK"],
                    "action": "1",
                    "expires": valid_datetime,
                }
            )
            form = IPv6Form(formdata=form_data)
            form.net_ranges = sample_network_ranges
            form.action.choices = mock_action_choices

            if not form.validate():
                print("Validation errors:", form.errors)

            assert form.validate()

    @pytest.mark.parametrize("port_data", ["80", "80;443", "1024-2048"])
    def test_valid_ports(self, app, valid_datetime, sample_network_ranges, mock_action_choices, port_data):
        """Test validation with various valid port formats"""
        with app.test_request_context():
            form_data = create_form_data(
                {
                    "source": "2001:db8::",
                    "source_mask": "32",
                    "next_header": "tcp",
                    "source_port": port_data,
                    "dest_port": port_data,
                    "action": "1",
                    "expires": valid_datetime,
                }
            )
            form = IPv6Form(formdata=form_data)
            form.net_ranges = sample_network_ranges
            form.action.choices = mock_action_choices

            if not form.validate():
                print("Validation errors:", form.errors)

            assert form.validate()


class TestBulkUserForm:
    @pytest.fixture
    def valid_csv_data(self):
        return "uuid-eppn,role,organizace\nuser1@example.com,2,1\nuser2@example.com,2,1"

    def test_valid_bulk_import(self, app, valid_csv_data):
        with app.test_request_context():
            form_data = create_form_data({"users": valid_csv_data})
            form = BulkUserForm(formdata=form_data)
            form.roles = {2}  # Mock available roles
            form.organizations = {1}  # Mock available organizations
            form.uuids = set()  # Mock existing UUIDs (empty set)
            assert form.validate()

    def test_duplicate_uuid(self, app, valid_csv_data):
        with app.test_request_context():
            form_data = create_form_data({"users": valid_csv_data})
            form = BulkUserForm(formdata=form_data)
            form.roles = {2}
            form.organizations = {1}
            form.uuids = {"user1@example.com"}  # UUID already exists
            assert not form.validate()
            assert any("already exists" in error for error in form.users.errors)

    def test_invalid_role(self, app):
        csv_data = "uuid-eppn,role,organizace\nuser1@example.com,999,1"  # Invalid role ID
        with app.test_request_context():
            form_data = create_form_data({"users": csv_data})
            form = BulkUserForm(formdata=form_data)
            form.roles = {2}
            form.organizations = {1}
            form.uuids = set()
            assert not form.validate()
            assert any("does not exist" in error for error in form.users.errors)


class TestApiKeyForm:
    @pytest.fixture
    def valid_api_key_data(self, valid_datetime):
        return {"machine": "192.168.1.1", "comment": "Test API key", "expires": valid_datetime, "readonly": "true"}

    def test_valid_api_key(self, app, valid_api_key_data):
        with app.test_request_context():
            form_data = create_form_data(valid_api_key_data)
            form = ApiKeyForm(formdata=form_data)
            assert form.validate()

    def test_invalid_ip(self, app, valid_datetime):
        with app.test_request_context():
            form_data = create_form_data({"machine": "invalid_ip", "expires": valid_datetime})
            form = ApiKeyForm(formdata=form_data)
            assert not form.validate()
            assert "provide valid IP address" in form.machine.errors

    def test_unlimited_expiration(self, app):
        with app.test_request_context():
            form_data = create_form_data({"machine": "192.168.1.1", "expires": ""})  # Empty expiration for unlimited
            form = ApiKeyForm(formdata=form_data)
            assert form.validate()


class TestMachineApiKeyForm:
    # Similar to ApiKeyForm, but might have different validation rules
    @pytest.fixture
    def valid_machine_key_data(self, valid_datetime):
        return {
            "machine": "192.168.1.1",
            "comment": "Test machine API key",
            "expires": valid_datetime,
            "readonly": "true",
            "user": 1,
        }

    def test_valid_machine_key(self, app, valid_machine_key_data):
        with app.test_request_context():
            form_data = create_form_data(valid_machine_key_data)
            form = MachineApiKeyForm(formdata=form_data)
            form.user.choices = [(1, "g.name"), (2, "test")]
            assert form.validate()


class TestOrganizationForm:
    @pytest.fixture
    def valid_org_data(self):
        return {
            "name": "Test Organization",
            "limit_flowspec4": "100",
            "limit_flowspec6": "100",
            "limit_rtbh": "50",
            "arange": "192.168.0.0/16\n2001:db8::/32",
        }

    def test_valid_organization(self, app, valid_org_data):
        with app.test_request_context():
            form_data = create_form_data(valid_org_data)
            form = OrganizationForm(formdata=form_data)
            assert form.validate()

    def test_invalid_ranges(self, app):
        with app.test_request_context():
            form_data = create_form_data({"name": "Test Org", "arange": "invalid_range\n192.168.0.0/16"})
            form = OrganizationForm(formdata=form_data)
            assert not form.validate()

    def test_invalid_limits(self, app):
        with app.test_request_context():
            form_data = create_form_data({"name": "Test Org", "limit_flowspec4": "1001"})  # Exceeds max value
            form = OrganizationForm(formdata=form_data)
            assert not form.validate()


class TestActionForm:
    @pytest.fixture
    def valid_action_data(self):
        return {
            "name": "test_action",
            "command": "announce route",
            "description": "Test action description",
            "role_id": "2",
        }

    def test_valid_action(self, app, valid_action_data):
        with app.test_request_context():
            form_data = create_form_data(valid_action_data)
            form = ActionForm(formdata=form_data)
            assert form.validate()

    def test_invalid_role(self, app, valid_action_data):
        with app.test_request_context():
            invalid_data = dict(valid_action_data)
            invalid_data["role_id"] = "4"  # Invalid role
            form_data = create_form_data(invalid_data)
            form = ActionForm(formdata=form_data)
            assert not form.validate()


class TestASPathForm:
    @pytest.fixture
    def valid_aspath_data(self):
        return {"prefix": "AS64512", "as_path": "64512 64513 64514"}

    def test_valid_aspath(self, app, valid_aspath_data):
        with app.test_request_context():
            form_data = create_form_data(valid_aspath_data)
            form = ASPathForm(formdata=form_data)
            assert form.validate()

    def test_empty_fields(self, app):
        with app.test_request_context():
            form_data = create_form_data({})
            form = ASPathForm(formdata=form_data)
            assert not form.validate()


class TestCommunityForm:
    @pytest.fixture
    def valid_community_data(self):
        return {
            "name": "test_community",
            "comm": "64512:100",
            "description": "Test community",
            "role_id": "2",
            "as_path": "true",
        }

    def test_valid_community(self, app, valid_community_data):
        with app.test_request_context():
            form_data = create_form_data(valid_community_data)
            form = CommunityForm(formdata=form_data)
            assert form.validate()

    def test_missing_all_community_types(self, app):
        with app.test_request_context():
            form_data = create_form_data({"name": "test_community", "role_id": "2"})
            form = CommunityForm(formdata=form_data)
            assert not form.validate()
            assert any("could not be empty" in error for error in form.comm.errors)

    def test_valid_with_extended_community(self, app):
        with app.test_request_context():
            form_data = create_form_data({"name": "test_community", "extcomm": "rt:64512:100", "role_id": "2"})
            form = CommunityForm(formdata=form_data)
            assert form.validate()

    def test_valid_with_large_community(self, app):
        with app.test_request_context():
            form_data = create_form_data({"name": "test_community", "larcomm": "64512:100:200", "role_id": "2"})
            form = CommunityForm(formdata=form_data)
            assert form.validate()


class TestWhitelistForm:
    @pytest.fixture
    def valid_ipv4_data(self, valid_datetime):
        return {"ip": "192.168.0.0", "mask": "24", "comment": "Test whitelist entry", "expires": valid_datetime}

    @pytest.fixture
    def valid_ipv6_data(self, valid_datetime):
        return {"ip": "2001:db8::", "mask": "32", "comment": "IPv6 whitelist entry", "expires": valid_datetime}

    def test_valid_ipv4_entry(self, app, valid_ipv4_data, sample_network_ranges):
        with app.test_request_context():
            form_data = create_form_data(valid_ipv4_data)
            form = WhitelistForm(formdata=form_data)
            form.net_ranges = sample_network_ranges

            if not form.validate():
                print("Validation errors:", form.errors)

            assert form.validate()

    def test_valid_ipv6_entry(self, app, valid_ipv6_data, sample_network_ranges):
        with app.test_request_context():
            form_data = create_form_data(valid_ipv6_data)
            form = WhitelistForm(formdata=form_data)
            form.net_ranges = sample_network_ranges

            if not form.validate():
                print("Validation errors:", form.errors)

            assert form.validate()

    def test_invalid_ip_format(self, app, valid_datetime, sample_network_ranges):
        with app.test_request_context():
            form_data = create_form_data({"ip": "invalid_ip", "mask": "24", "expires": valid_datetime})
            form = WhitelistForm(formdata=form_data)
            form.net_ranges = sample_network_ranges
            assert not form.validate()
            assert any("valid IP address" in error for error in form.ip.errors)

    def test_ip_outside_range(self, app, valid_datetime):
        with app.test_request_context():
            form_data = create_form_data(
                {"ip": "10.0.0.0", "mask": "24", "expires": valid_datetime}  # IP outside allowed range
            )
            form = WhitelistForm(formdata=form_data)
            form.net_ranges = ["192.168.0.0/16"]  # Only allow 192.168.0.0/16
            assert not form.validate()
            assert any("must be in organization range" in error for error in form.ip.errors)

    def test_invalid_mask_ipv4(self, app, valid_datetime, sample_network_ranges):
        with app.test_request_context():
            form_data = create_form_data(
                {"ip": "192.168.0.0", "mask": "33", "expires": valid_datetime}  # Invalid mask for IPv4
            )
            form = WhitelistForm(formdata=form_data)
            form.net_ranges = sample_network_ranges
            assert not form.validate()

    def test_invalid_mask_ipv6(self, app, valid_datetime, sample_network_ranges):
        with app.test_request_context():
            form_data = create_form_data(
                {"ip": "2001:db8::", "mask": "129", "expires": valid_datetime}  # Invalid mask for IPv6
            )
            form = WhitelistForm(formdata=form_data)
            form.net_ranges = sample_network_ranges
            assert not form.validate()

    def test_missing_required_fields(self, app, sample_network_ranges):
        with app.test_request_context():
            form_data = create_form_data({})
            form = WhitelistForm(formdata=form_data)
            form.net_ranges = sample_network_ranges
            assert not form.validate()
            assert form.ip.errors  # Should have error for missing IP
            assert form.mask.errors  # Should have error for missing mask
            assert form.expires.errors  # Should have error for missing expiration

    def test_comment_length(self, app, valid_datetime, sample_network_ranges):
        with app.test_request_context():
            form_data = create_form_data(
                {
                    "ip": "192.168.0.0",
                    "mask": "24",
                    "expires": valid_datetime,
                    "comment": "x" * 256,  # Comment longer than 255 chars
                }
            )
            form = WhitelistForm(formdata=form_data)
            form.net_ranges = sample_network_ranges
            assert not form.validate()

    @pytest.mark.parametrize(
        "expires", ["", "invalid_date", "2020-33-42T00:00"]  # Empty expiration  # Invalid date format  # Past date
    )
    def test_invalid_expiration(self, app, expires, sample_network_ranges):
        with app.test_request_context():
            form_data = create_form_data({"ip": "192.168.0.0", "mask": "24", "expires": expires})
            form = WhitelistForm(formdata=form_data)
            form.net_ranges = sample_network_ranges
            if not form.validate():
                print("Validation errors:", form.errors)

            assert not form.validate()

    def test_network_alignment(self, app, valid_datetime, sample_network_ranges):
        """Test that IP addresses must be properly network-aligned"""
        with app.test_request_context():
            form_data = create_form_data(
                {"ip": "192.168.1.1", "mask": "24", "expires": valid_datetime}  # Not aligned to /24 boundary
            )
            form = WhitelistForm(formdata=form_data)
            form.net_ranges = sample_network_ranges
            assert not form.validate()



from datetime import datetime, timedelta
from flask import flash
from flowapp.constants import (
    COMP_FUNCS,
    TIME_YEAR,
    TIME_US,
    TIME_STMP,
    TIME_FORMAT_ARG,
    RULE_TYPES_DICT,
    FORM_TIME_PATTERN,
)


def other_rtypes(rtype):
    """
    get rtype and return list of remaining rtypes
    for example get ipv4 and return [ipv6, rtbh]
    """
    result = list(RULE_TYPES_DICT.keys())
    try:
        result.remove(rtype)
    except ValueError:
        pass

    return result


def output_date_format(json_request_data, pref_format=TIME_YEAR):
    """
    prefer user setting from parameter, if the parameter is not set
    then use the prefered format computed from input date
    """
    if not json_request_data:
        return pref_format

    if TIME_FORMAT_ARG in json_request_data and json_request_data[TIME_FORMAT_ARG]:
        return json_request_data[TIME_FORMAT_ARG]
    else:
        return pref_format


def parse_api_time(apitime):
    """
    check if the api time is in US, EU or timestamp format
    :param apitime: string with date and time
    :returns: datetime, prefered format
    """

    apitime = str(apitime)
    try:
        return (
            round_to_ten_minutes(datetime.strptime(apitime, FORM_TIME_PATTERN)),
            TIME_US,
        )
    except ValueError:
        mytime = False

    try:
        return round_to_ten_minutes(webpicker_to_datetime(apitime)), TIME_YEAR
    except ValueError:
        mytime = False

    try:
        return round_to_ten_minutes(webpicker_to_datetime(apitime, TIME_US)), TIME_US
    except ValueError:
        mytime = False

    try:
        return round_to_ten_minutes(datetime.fromtimestamp(int(apitime))), TIME_STMP
    except OverflowError:
        mytime = False
    except ValueError:
        mytime = False

    return mytime


def quote_to_ent(comment):
    """
    Convert all " to &quot;
    Used for comment sanitize / because of tooltip in dashboard break when quotes are unescaped
    :param comment: string to be sanitized
    :return: string
    """
    if comment:
        return comment.replace('"', "&quot;")


def webpicker_to_datetime(webtime, format=TIME_YEAR):
    """
    convert 'YYYY/MM/DD HH:mm' to datetime
    """
    if format == TIME_YEAR:
        formating_string = "%Y/%m/%d %H:%M"
    else:
        formating_string = "%m/%d/%Y %H:%M"

    return datetime.strptime(webtime, formating_string)


def datetime_to_webpicker(python_time, format=TIME_YEAR):
    """
    convert datetime to 'YYYY/MM/DD HH:mm' string
    """
    if format == TIME_YEAR:
        formating_string = "%Y/%m/%d %H:%M"
    else:
        formating_string = "%m/%d/%Y %H:%M"

    return datetime.strftime(python_time, formating_string)


def get_state_by_time(python_time):
    """
    returns state for rule based on given time
    if given time is in the past returns 2 (withdrawed rule)
    else returns 1
    :param python_time:
    :return: integer rstate
    """
    present = datetime.now()

    if python_time <= present:
        return 2
    else:
        return 1


def round_to_ten_minutes(python_time):
    """
    Round given time to nearest ten minutes
    :param python_time: datetime
    :return: datetime
    """
    python_time += timedelta(minutes=5)
    python_time -= timedelta(
        minutes=python_time.minute % 10,
        seconds=python_time.second,
        microseconds=python_time.microsecond,
    )

    return python_time


def flash_errors(form):
    """
    Flash all error messages
    :param form: WTForm object
    :return: none
    """
    for field, errors in form.errors.items():
        for error in errors:
            flash("Error in the %s field - %s" % (getattr(form, field).label.text, error))


def active_css_rstate(rtype, rstate):
    """
    returns dict with rstates as keys and css class value
    :param rstate: string
    :return: dict
    """

    return {
        "active": "",
        "expired": "",
        "all": "",
        "ipv4": "",
        "ipv6": "",
        "rtbh": "",
        rtype: "active",
        rstate: "active",
    }


def get_comp_func(rstate="active"):
    try:
        comp_func = COMP_FUNCS[rstate]
    except IndexError:
        comp_func = None
    except KeyError:
        comp_func = None

    return comp_func



import json
from datetime import datetime, timedelta

from flowapp.models import RTBH
from flowapp.models.rules.whitelist import Whitelist


def test_create_rtbh_after_expired_rule_exists(client, app, db, jwt_token):
    """
    Test that demonstrates the bug: creating a new RTBH rule with the same IP
    as an expired rule results in the new rule having withdrawn state instead
    of active state.

    Test should be run in isolation or as the last in stack.

    Steps:
    1. Create an RTBH rule with expiration in the past (will be withdrawn, rstate_id=2)
    2. Create another RTBH rule with the same IP but expiration in the future
    3. Verify that the second rule should be active (rstate_id=1) but is actually withdrawn (rstate_id=2)
    """
    # cleanup any existing RTBH rules to avoid interference
    cleanup_before_stack(app, db)

    # Step 1: Create an expired RTBH rule
    expired_payload = {
        "community": 1,
        "ipv4": "192.168.100.50",
        "ipv4_mask": 32,
        "expires": (datetime.now() - timedelta(days=1)).strftime("%m/%d/%Y %H:%M"),
        "comment": "Expired rule that should be in withdrawn state",
    }

    response1 = client.post(
        "/api/v3/rules/rtbh",
        headers={"x-access-token": jwt_token},
        json=expired_payload,
    )

    assert response1.status_code == 201
    data1 = json.loads(response1.data)
    rule_id_1 = data1["rule"]["id"]

    # Verify the first rule is in withdrawn state
    with app.app_context():
        expired_rule = db.session.query(RTBH).filter_by(id=rule_id_1).first()
        assert expired_rule is not None
        assert expired_rule.rstate_id == 2, "Expired rule should be in withdrawn state (rstate_id=2)"
        assert expired_rule.ipv4 == "192.168.100.50"
        assert expired_rule.ipv4_mask == 32
        print(f"✓ First rule created with ID {rule_id_1}, state: {expired_rule.rstate_id} (withdrawn)")

    # Step 2: Create a new RTBH rule with the same IP but future expiration
    future_payload = {
        "community": 1,
        "ipv4": "192.168.100.50",
        "ipv4_mask": 32,
        "expires": (datetime.now() + timedelta(days=7)).strftime("%m/%d/%Y %H:%M"),
        "comment": "New rule that should be active but will be withdrawn due to bug",
    }

    response2 = client.post(
        "/api/v3/rules/rtbh",
        headers={"x-access-token": jwt_token},
        json=future_payload,
    )

    assert response2.status_code == 201
    data2 = json.loads(response2.data)
    rule_id_2 = data2["rule"]["id"]

    # Step 3: Verify the bug - the second rule should be active but is withdrawn
    with app.app_context():
        # The bug causes the expired rule to be updated instead of creating a new one
        # OR if a new rule is created, it has the wrong state

        # Check if it's the same rule (updated) or a new rule
        total_rules = db.session.query(RTBH).filter_by(ipv4="192.168.100.50", ipv4_mask=32).count()

        new_rule = db.session.query(RTBH).filter_by(id=rule_id_2).first()
        assert new_rule is not None

        print("\n--- Bug Verification ---")
        print(f"Total rules with IP 192.168.100.50/32: {total_rules}")
        print(f"First rule ID: {rule_id_1}")
        print(f"Second rule ID: {rule_id_2}")
        print(f"Same rule updated: {rule_id_1 == rule_id_2}")
        print(f"Second rule state: {new_rule.rstate_id}")
        print(f"Second rule expires: {new_rule.expires}")
        print(f"Expiration is in future: {new_rule.expires > datetime.now()}")

        # The bug: even though expiration is in the future, the rule is in withdrawn state
        # EXPECTED: rstate_id should be 1 (active)
        # ACTUAL: rstate_id is 2 (withdrawn) due to the bug

        # This assertion will FAIL due to the bug, demonstrating the issue
        assert new_rule.expires > datetime.now(), "Rule expiration should be in the future"

        # THIS IS THE BUG: The rule has future expiration but is in withdrawn state
        try:
            assert new_rule.rstate_id == 1, (
                f"BUG DETECTED: Rule with future expiration should be active (rstate_id=1), "
                f"but is in state {new_rule.rstate_id}. "
                f"This happens because the expired rule was found and updated without resetting the state."
            )
            print("✓ Test PASSED - bug is fixed!")
        except AssertionError as e:
            print(f"✗ Test FAILED - bug confirmed: {e}")
            raise
    cleanup_rtbh_rule(app, db, rule_id_1)
    cleanup_rtbh_rule(app, db, rule_id_2)


def test_create_rtbh_after_expired_rule_different_mask(client, app, db, jwt_token):
    """
    Test that verifies the bug only occurs when IP AND mask match.
    When the mask is different, a new rule should be created successfully.
    """

    # Step 1: Create an expired RTBH rule with /32 mask
    expired_payload = {
        "community": 1,
        "ipv4": "192.168.100.60",
        "ipv4_mask": 32,
        "expires": (datetime.now() - timedelta(days=1)).strftime("%m/%d/%Y %H:%M"),
        "comment": "Expired /32 rule",
    }

    response1 = client.post(
        "/api/v3/rules/rtbh",
        headers={"x-access-token": jwt_token},
        json=expired_payload,
    )

    assert response1.status_code == 201

    # Step 2: Create a new rule with same IP but different mask (/24)
    future_payload = {
        "community": 1,
        "ipv4": "192.168.100.0",
        "ipv4_mask": 24,
        "expires": (datetime.now() + timedelta(days=7)).strftime("%m/%d/%Y %H:%M"),
        "comment": "New /24 rule - should be active",
    }

    response2 = client.post(
        "/api/v3/rules/rtbh",
        headers={"x-access-token": jwt_token},
        json=future_payload,
    )

    assert response2.status_code == 201
    data2 = json.loads(response2.data)

    # Verify the new rule is active (this should work because IP+mask don't match)
    with app.app_context():
        new_rule = db.session.query(RTBH).filter_by(id=data2["rule"]["id"]).first()
        assert new_rule is not None
        assert new_rule.rstate_id == 1, "New rule with different mask should be active"
        print("✓ Different mask creates new active rule correctly")

    cleanup_rtbh_rule(app, db, new_rule.id)


def test_create_rtbh_after_active_rule_exists(client, app, db, jwt_token):
    """
    Test that when an active rule exists, updating it with a new expiration
    maintains the active state (this should work correctly).
    """

    # Step 1: Create an active RTBH rule
    active_payload = {
        "community": 1,
        "ipv4": "192.168.100.70",
        "ipv4_mask": 32,
        "expires": (datetime.now() + timedelta(days=1)).strftime("%m/%d/%Y %H:%M"),
        "comment": "Active rule",
    }

    response1 = client.post(
        "/api/v3/rules/rtbh",
        headers={"x-access-token": jwt_token},
        json=active_payload,
    )

    assert response1.status_code == 201
    data1 = json.loads(response1.data)
    rule_id_1 = data1["rule"]["id"]

    # Verify the first rule is active
    with app.app_context():
        first_rule = db.session.query(RTBH).filter_by(id=rule_id_1).first()
        assert first_rule.rstate_id == 1, "First rule should be active"

    # Step 2: Update the same rule with a new expiration
    updated_payload = {
        "community": 1,
        "ipv4": "192.168.100.70",
        "ipv4_mask": 32,
        "expires": (datetime.now() + timedelta(days=7)).strftime("%m/%d/%Y %H:%M"),
        "comment": "Updated active rule",
    }

    response2 = client.post(
        "/api/v3/rules/rtbh",
        headers={"x-access-token": jwt_token},
        json=updated_payload,
    )

    assert response2.status_code == 201
    data2 = json.loads(response2.data)

    # Verify it maintains active state
    with app.app_context():
        updated_rule = db.session.query(RTBH).filter_by(id=data2["rule"]["id"]).first()
        assert updated_rule is not None
        assert updated_rule.rstate_id == 1, "Updated rule should remain active"
        print("✓ Updating active rule maintains active state correctly")

    cleanup_rtbh_rule(app, db, rule_id_1)


def cleanup_before_stack(app, db):
    """
    Cleanup function to remove all RTBH rules created during tests.
    """
    with app.app_context():
        db.session.query(RTBH).delete()
        db.session.query(Whitelist).delete()
        db.session.commit()


def cleanup_rtbh_rule(app, db, rule_id):
    """
    Cleanup function to remove RTBH rule created during tests.
    """
    with app.app_context():
        rule = db.session.get(RTBH, rule_id)
        if rule:
            db.session.delete(rule)
            db.session.commit()



from .rule_service import (
    create_or_update_ipv4_rule,
    create_or_update_ipv6_rule,
    create_or_update_rtbh_rule,
    reactivate_rule,
)

from .whitelist_service import create_or_update_whitelist, delete_whitelist, delete_expired_whitelists

from .base import announce_all_routes

__all__ = [
    create_or_update_ipv4_rule,
    create_or_update_ipv6_rule,
    create_or_update_rtbh_rule,
    create_or_update_whitelist,
    delete_whitelist,
    delete_expired_whitelists,
    announce_all_routes,
    reactivate_rule,
]



from datetime import datetime, timedelta

from flowapp.constants import RuleTypes
from .base import db


class Log(db.Model):
    """Log model for system actions"""

    id = db.Column(db.Integer, primary_key=True)
    time = db.Column(db.DateTime)
    task = db.Column(db.String(1000))
    author = db.Column(db.String(1000))
    rule_type = db.Column(db.Integer)
    rule_id = db.Column(db.Integer)
    user_id = db.Column(db.Integer)

    def __init__(self, time, task, user_id, rule_type, rule_id, author):
        self.time = time
        self.task = task
        self.rule_type = rule_type
        self.rule_id = rule_id
        self.user_id = user_id
        self.author = author

    @classmethod
    def delete_old(cls, days: int = 30):
        """Delete logs older than :param days from the database"""
        cls.query.filter(cls.time < datetime.now() - timedelta(days=days)).delete()
        db.session.commit()

    def __repr__(self):
        return f"<Log {self.id}>"

    def __str__(self):
        """
        {"author": "vrany@cesnet.cz / Cel\u00fd sv\u011bt", "source": "UI", "command": "cmd"}
        """
        return f"{self.author} - {RuleTypes(self.rule_type).name}({self.rule_id}) - {self.task}"



import logging
import babel
from flask import redirect, render_template, request, session, url_for


def register_blueprints(app, csrf=None):
    """Register Flask blueprints."""
    from flowapp.views.admin import admin
    from flowapp.views.rules import rules
    from flowapp.views.api_v1 import api as api_v1
    from flowapp.views.api_v2 import api as api_v2
    from flowapp.views.api_v3 import api as api_v3
    from flowapp.views.api_keys import api_keys
    from flowapp.views.dashboard import dashboard
    from flowapp.views.whitelist import whitelist

    # Configure CSRF exemption for API routes
    if csrf:
        csrf.exempt(api_v1)
        csrf.exempt(api_v2)
        csrf.exempt(api_v3)

    # Register blueprints with URL prefixes
    app.register_blueprint(admin, url_prefix="/admin")
    app.register_blueprint(rules, url_prefix="/rules")
    app.register_blueprint(api_keys, url_prefix="/api_keys")
    app.register_blueprint(api_v1, url_prefix="/api/v1")
    app.register_blueprint(api_v2, url_prefix="/api/v2")
    app.register_blueprint(api_v3, url_prefix="/api/v3")
    app.register_blueprint(dashboard, url_prefix="/dashboard")
    app.register_blueprint(whitelist, url_prefix="/whitelist")

    return app


def configure_logging(app):
    """Configure logging for the Flask application."""

    # Remove all default handlers
    for handler in app.logger.handlers[:]:
        app.logger.removeHandler(handler)

    # Retrieve log level and file name from config
    log_level = app.config.get("LOG_LEVEL", "DEBUG").upper()
    log_file = app.config.get("LOG_FILE", "app.log")

    # Define log format
    log_format = "%(asctime)s | %(levelname)s | %(message)s"
    log_datefmt = "%Y-%m-%d %H:%M:%S"
    formatter = logging.Formatter(log_format, datefmt=log_datefmt)

    # Console handler
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(formatter)

    # File handler
    file_handler = logging.FileHandler(log_file)
    file_handler.setFormatter(formatter)

    # Set logger level
    app.logger.setLevel(getattr(logging, log_level, logging.DEBUG))

    # Attach handlers
    app.logger.addHandler(console_handler)
    app.logger.addHandler(file_handler)

    return app


def register_error_handlers(app):
    """Register error handlers."""

    @app.errorhandler(404)
    def not_found(error):
        return render_template("errors/404.html"), 404

    @app.errorhandler(500)
    def internal_error(exception):
        app.logger.exception(exception)
        return render_template("errors/500.html"), 500

    return app


def register_context_processors(app):
    """Register template context processors."""

    @app.context_processor
    def utility_processor():
        def editable_rule(rule):
            if rule:
                from flowapp.validators import editable_range
                from flowapp.models import get_user_nets

                editable_range(rule, get_user_nets(session["user_id"]))
                return True
            return False

        return dict(editable_rule=editable_rule)

    @app.context_processor
    def inject_main_menu():
        """Inject main menu config to templates."""
        return {"main_menu": app.config.get("MAIN_MENU")}

    @app.context_processor
    def inject_dashboard():
        """Inject dashboard config to templates."""
        return {"dashboard": app.config.get("DASHBOARD")}

    @app.context_processor
    def inject_footer_menu():
        """Inject main menu config to templates."""
        return {"footer_menu": app.config.get("FOOTER_MENU", [])}

    return app


def register_template_filters(app):
    """Register custom template filters."""

    @app.template_filter("strftime")
    def format_datetime(value):
        if value is None:
            return app.config.get("MISSING_DATETIME_MESSAGE", "Never")

        format = "y/MM/dd HH:mm"
        locale = app.config.get("BABEL_DEFAULT_LOCALE", "en_US_POSIX")
        return babel.dates.format_datetime(value, format, locale=locale)

    @app.template_filter("unlimited")
    def unlimited_filter(value):
        return "unlimited" if value == 0 else value

    return app


def register_auth_handlers(app, ext):
    """Register authentication handlers."""

    @ext.login_handler
    def login(user_info):
        try:
            uuid = user_info.get("eppn")
        except KeyError:
            uuid = False
            return render_template("errors/401.html")

        return _handle_login(uuid, app)

    @app.route("/logout")
    def logout():
        session["user_uuid"] = False
        session["user_id"] = False
        session.clear()
        return redirect(app.config.get("LOGOUT_URL"))

    @app.route("/ext-login")
    def ext_login():
        header_name = app.config.get("AUTH_HEADER_NAME", "X-Authenticated-User")
        if header_name not in request.headers:
            return render_template("errors/401.html")

        uuid = request.headers.get(header_name)
        if not uuid:
            return render_template("errors/401.html")

        return _handle_login(uuid, app)

    @app.route("/local-login")
    def local_login():
        print("Local login started")
        if not app.config.get("LOCAL_AUTH", False):
            print("Local auth not enabled")
            return render_template("errors/401.html")

        uuid = app.config.get("LOCAL_USER_UUID", False)
        if not uuid:
            print("Local user not set")
            return render_template("errors/401.html")

        print(f"Local login with {uuid}")
        return _handle_login(uuid, app)

    return app


def _handle_login(uuid, app):
    """Handle login process for all authentication methods."""
    from flowapp import db

    multiple_orgs = False
    try:
        user, multiple_orgs = _register_user_to_session(uuid, db)
    except AttributeError as e:
        app.logger.exception(e)
        return render_template("errors/401.html")

    if multiple_orgs:
        return redirect(url_for("select_org", org_id=None))

    # set user org to session
    user_org = user.organization.first()
    session["user_org"] = user_org.name
    session["user_org_id"] = user_org.id

    return redirect("/")


def _register_user_to_session(uuid, db):
    """Register user information to session."""
    from flowapp.models import User

    print(f"Registering user {uuid} to session")
    user = db.session.query(User).filter_by(uuid=uuid).first()
    print(f"Got user {user} from DB")
    session["user_uuid"] = user.uuid
    session["user_email"] = user.uuid
    session["user_name"] = user.name
    session["user_id"] = user.id
    session["user_roles"] = [role.name for role in user.role.all()]
    session["user_role_ids"] = [role.id for role in user.role.all()]
    roles = [i > 1 for i in session["user_role_ids"]]
    session["can_edit"] = True if all(roles) and roles else []
    # check if user has multiple organizations and return True if so
    return user, len(user.organization.all()) > 1



"""
Base rule form for the flowapp application.
"""

from flask_wtf import FlaskForm
from wtforms import SelectMultipleField, StringField, SelectField, TextAreaField
from wtforms.validators import Optional, Length, DataRequired, InputRequired

from ..base import MultiFormatDateTimeLocalField
from ...constants import TCP_FLAGS
from ...validators import PortString, address_with_mask, network_in_range, whole_world_range


class IPForm(FlaskForm):
    """
    Base class for IPv4 and IPv6 rules
    """

    def __init__(self, *args, **kwargs):
        super(IPForm, self).__init__(*args, **kwargs)
        self.net_ranges = None

    zero_address = None
    source = None
    source_mask = None
    dest = None
    dest_mask = None
    flags = SelectMultipleField("TCP flag(s)", choices=TCP_FLAGS, validators=[Optional()])

    source_port = StringField(
        "Source port(s) -  ; separated ",
        validators=[Optional(), Length(max=255), PortString()],
    )

    dest_port = StringField(
        "Destination port(s) - ; separated",
        validators=[Optional(), Length(max=255), PortString()],
    )

    packet_len = StringField(
        "Packet length - ; separated ",
        validators=[Optional(), Length(max=255), PortString()],
    )

    action = SelectField(
        "Action",
        coerce=int,
        validators=[DataRequired(message="Please select an action for the rule.")],
    )

    expires = MultiFormatDateTimeLocalField("Expires", format="%Y-%m-%dT%H:%M", validators=[InputRequired()])

    comment = arange = TextAreaField("Comments")

    def validate(self):
        """
        custom validation method
        :return: boolean
        """

        result = True
        if not FlaskForm.validate(self):
            result = False

        source = self.validate_source_address()
        dest = self.validate_dest_address()
        ranges = self.validate_address_ranges()
        ips = self.validate_ipv_specific()

        return result and source and dest and ranges and ips

    def validate_source_address(self):
        """
        validate source address, set error message if validation fails
        :return: boolean validation result
        """
        if self.source.data and not address_with_mask(self.source.data, self.source_mask.data):
            self.source.errors.append(
                "This is not valid combination of address {} and mask {}.".format(
                    self.source.data, self.source_mask.data
                )
            )
            return False

        return True

    def validate_dest_address(self):
        """
        validate dest address, set error message if validation fails
        :return: boolean validation result
        """
        if self.dest.data and not address_with_mask(self.dest.data, self.dest_mask.data):
            self.dest.errors.append(
                "This is not valid combination of address {} and mask {}.".format(self.dest.data, self.dest_mask.data)
            )
            return False

        return True

    def validate_address_ranges(self):
        """
        validates if the address of source is in the user range
        if the source and dest address are empty, check if the user
        is member of whole world organization
        :return: boolean validation result
        """
        if not (self.source.data or self.dest.data):
            whole_world_member = whole_world_range(self.net_ranges, self.zero_address)
            if not whole_world_member:
                self.source.errors.append("Source or dest must be in organization range : {}.".format(self.net_ranges))
                self.dest.errors.append("Source or dest must be in organization range : {}.".format(self.net_ranges))
                return False
        else:
            source_in_range = network_in_range(self.source.data, self.source_mask.data, self.net_ranges)
            dest_in_range = network_in_range(self.dest.data, self.dest_mask.data, self.net_ranges)
            if not (source_in_range or dest_in_range):
                self.source.errors.append("Source or dest must be in organization range : {}.".format(self.net_ranges))
                self.dest.errors.append("Source or dest must be in organization range : {}.".format(self.net_ranges))
                return False

        return True

    def validate_ipv_specific(self):
        """
        abstract method must be implemented in the subclass
        """
        pass



import json


from flowapp.models import Flowspec4, Organization

V_PREFIX = "/api/v3"


def test_create_rtbh_only(client, app, db, jwt_token):
    """
    Test creating an RTBH rule through API that exactly matches an existing whitelist.

    The rule should be created but marked as whitelisted (rstate_id=4), and a cache entry
    should be created linking the rule to the whitelist.
    """

    # Now create the RTBH rule via API
    res = client.post(
        f"{V_PREFIX}/rules/rtbh",
        headers={"x-access-token": jwt_token},
        json={
            "community": 1,
            "ipv4": "147.230.17.17",
            "ipv4_mask": 32,
            "expires": "10/25/2050 14:46",
        },
    )

    # Verify response is successful
    assert res.status_code == 201
    data = json.loads(res.data)
    assert data["rule"] is not None


def test_token(client, jwt_token):
    """
    test that token authorization works
    """
    req = client.get(f"{V_PREFIX}/test_token", headers={"x-access-token": jwt_token})

    assert req.status_code == 200


def test_withnout_token(client):
    """
    test that without token authorization return 401
    """
    req = client.get(f"{V_PREFIX}/test_token")

    assert req.status_code == 401


def test_list_actions(client, db, jwt_token):
    """
    test that endpoint returns all action in db
    """
    req = client.get(f"{V_PREFIX}/actions", headers={"x-access-token": jwt_token})

    assert req.status_code == 200
    data = json.loads(req.data)
    assert len(data) == 4


def test_list_communities(client, db, jwt_token):
    """
    test that endpoint returns all action in db
    """
    req = client.get(f"{V_PREFIX}/communities", headers={"x-access-token": jwt_token})

    assert req.status_code == 200
    data = json.loads(req.data)
    assert len(data) == 3


def test_create_v4rule(client, db, jwt_token):
    """
    test that creating with valid data returns 201
    """
    req = client.post(
        f"{V_PREFIX}/rules/ipv4",
        headers={"x-access-token": jwt_token},
        json={
            "action": 2,
            "protocol": "tcp",
            "source": "147.230.17.17",
            "source_mask": 32,
            "source_port": "",
            "expires": "10/15/2050 14:46",
            "flags": ["SYN", "RST"],
        },
    )

    assert req.status_code == 201
    data = json.loads(req.data)
    assert data["rule"]
    assert data["rule"]["id"] == 1
    assert data["rule"]["user"] == "jiri.vrany@cesnet.cz"


def test_delete_v4rule(client, db, jwt_token):
    """
    test that creating with valid data returns 201
    that time in the past creates expired rule (state 2)
    and that the rule deletion works as expected
    """
    req = client.post(
        f"{V_PREFIX}/rules/ipv4",
        headers={"x-access-token": jwt_token},
        json={
            "action": 2,
            "protocol": "tcp",
            "source": "147.230.17.12",
            "source_mask": 32,
            "source_port": "",
            "expires": "10/15/2015 14:46",
        },
    )

    assert req.status_code == 201
    data = json.loads(req.data)
    assert data["rule"]["id"] == 2
    assert data["rule"]["rstate"] == "withdrawed rule"

    req2 = client.delete(
        f'{V_PREFIX}/rules/ipv4/{data["rule"]["id"]}',
        headers={"x-access-token": jwt_token},
    )
    assert req2.status_code == 201


def test_create_rtbh_rule(client, db, jwt_token):
    """
    test that creating with valid data returns 201
    """
    req = client.post(
        f"{V_PREFIX}/rules/rtbh",
        headers={"x-access-token": jwt_token},
        json={
            "community": 1,
            "ipv4": "147.230.17.17",
            "ipv4_mask": 32,
            "expires": "10/25/2050 14:46",
        },
    )
    data = json.loads(req.data)
    assert req.status_code == 201
    assert data["rule"]
    assert data["rule"]["id"] == 1
    assert data["rule"]["user"] == "jiri.vrany@cesnet.cz"


def test_delete_rtbh_rule(client, db, jwt_token):
    """
    test that creating with valid data returns 201
    """
    req = client.post(
        f"{V_PREFIX}/rules/rtbh",
        headers={"x-access-token": jwt_token},
        json={
            "community": 2,
            "ipv4": "147.230.17.177",
            "ipv4_mask": 32,
            "expires": "10/25/2050 14:46",
        },
    )

    assert req.status_code == 201
    data = json.loads(req.data)
    assert data["rule"]["id"] == 2
    req2 = client.delete(
        f'{V_PREFIX}/rules/rtbh/{data["rule"]["id"]}',
        headers={"x-access-token": jwt_token},
    )
    assert req2.status_code == 201


def test_validation_rtbh_rule(client, db, jwt_token):
    """
    test that creating with invalid data returns 400 and errors
    """
    req = client.post(
        f"{V_PREFIX}/rules/rtbh",
        headers={"x-access-token": jwt_token},
        json={
            "community": 1,
            "ipv4": "147.230.17.17",
            "ipv4_mask": 32,
            "ipv6": "2001:718:1C01:1111::",
            "ipv6_mask": 128,
            "expires": "10/25/2050 14:46",
        },
    )
    data = json.loads(req.data)
    assert req.status_code == 400
    assert data["message"] == "error - invalid request data"
    assert type(data["validation_errors"]) is dict
    assert "ipv6" in data["validation_errors"]
    assert "ipv4" in data["validation_errors"]


def test_create_v6rule(client, db, jwt_token):
    """
    test that creating with valid data returns 201
    """
    req = client.post(
        f"{V_PREFIX}/rules/ipv6",
        headers={"x-access-token": jwt_token},
        json={
            "action": 3,
            "next_header": "tcp",
            "source": "2001:718:1C01:1111::",
            "source_mask": 64,
            "source_port": "",
            "expires": "10/25/2050 14:46",
        },
    )
    data = json.loads(req.data)
    assert req.status_code == 201
    assert data["rule"]
    assert data["rule"]["id"] == "1"
    assert data["rule"]["user"] == "jiri.vrany@cesnet.cz"


def test_validation_v4rule(client, db, jwt_token):
    """
    test that creating with invalid data returns 400 and errors
    """
    req = client.post(
        f"{V_PREFIX}/rules/ipv4",
        headers={"x-access-token": jwt_token},
        json={
            "action": 2,
            "dest": "200.200.200.32",
            "dest_mask": 16,
            "protocol": "tcp",
            "source": "1.1.1.1",
            "source_mask": 32,
            "source_port": "",
            "expires": "10/15/2050 14:46",
        },
    )

    assert req.status_code == 400
    data = json.loads(req.data)
    assert len(data["validation_errors"]) > 0
    assert sorted(data["validation_errors"].keys()) == sorted(["dest", "source"])
    assert len(data["validation_errors"]["dest"]) == 2
    assert data["validation_errors"]["dest"][0].startswith("This is not")
    assert data["validation_errors"]["dest"][1].startswith("Source or des")
    assert len(data["validation_errors"]["source"]) == 1
    assert data["validation_errors"]["source"][0].startswith("Source or des")


def test_all_validation_errors(client, db, jwt_token):
    """
    test that creating with invalid data returns 400 and errors
    """
    req = client.post(f"{V_PREFIX}/rules/ipv4", headers={"x-access-token": jwt_token}, json={"action": 2})
    assert req.status_code == 400


def test_validate_v6rule(client, db, jwt_token):
    """
    test that creating with invalid data returns 400 and errors
    """
    req = client.post(
        f"{V_PREFIX}/rules/ipv6",
        headers={"x-access-token": jwt_token},
        json={
            "action": 32,
            "next_header": "abc",
            "source": "2011:78:1C01:1111::",
            "source_mask": 64,
            "source_port": "",
            "expires": "10/25/2050 14:46",
        },
    )
    data = json.loads(req.data)
    assert req.status_code == 400
    assert len(data["validation_errors"]) > 0
    assert sorted(data["validation_errors"].keys()) == sorted(["action", "next_header", "dest", "source"])
    # assert data['validation_errors'][0].startswith('Error in the Action')
    # assert data['validation_errors'][1].startswith('Error in the Source')
    # assert data['validation_errors'][2].startswith('Error in the Next Header')


def test_rules(client, db, jwt_token):
    """
    test that there is one ipv4 rule created in the first test
    """
    req = client.get(f"{V_PREFIX}/rules", headers={"x-access-token": jwt_token})

    assert req.status_code == 200

    data = json.loads(req.data)
    assert len(data["flowspec_ipv4_rw"]) == 1
    assert len(data["flowspec_ipv6_rw"]) == 1


def test_timestamp_param(client, db, jwt_token):
    """
    test that url param for time format works as expected
    """
    req = client.get(f"{V_PREFIX}/rules?time_format=timestamp", headers={"x-access-token": jwt_token})

    assert req.status_code == 200

    data = json.loads(req.data)
    assert data["flowspec_ipv4_rw"][0]["expires"] == 2549451000
    assert data["flowspec_ipv6_rw"][0]["expires"] == 2550315000


def test_update_existing_v4rule_with_timestamp(client, db, jwt_token):
    """
    test that update with different data passes
    """
    req = client.post(
        f"{V_PREFIX}/rules/ipv4",
        headers={"x-access-token": jwt_token},
        json={
            "action": 2,
            "protocol": "tcp",
            "source": "147.230.17.17",
            "source_mask": 32,
            "source_port": "",
            "expires": "1444913400",
        },
    )

    assert req.status_code == 201
    data = json.loads(req.data)
    assert data["rule"]
    assert data["rule"]["id"] == 2
    assert data["rule"]["user"] == "jiri.vrany@cesnet.cz"
    assert data["rule"]["expires"] == 1444913400


def test_create_v4rule_with_timestamp(client, db, jwt_token):
    """
    test that creating with valid data returns 201
    """
    req = client.post(
        f"{V_PREFIX}/rules/ipv4",
        headers={"x-access-token": jwt_token},
        json={
            "action": 2,
            "protocol": "tcp",
            "source": "147.230.17.117",
            "source_mask": 32,
            "source_port": "",
            "expires": "1444913400",
        },
    )

    assert req.status_code == 201
    data = json.loads(req.data)
    assert data["rule"]
    assert data["rule"]["id"] == 3
    assert data["rule"]["user"] == "jiri.vrany@cesnet.cz"
    assert data["rule"]["expires"] == 1444913400


def test_update_existing_v6rule_with_timestamp(client, db, jwt_token):
    """
    test that update with different data passes
    """
    req = client.post(
        f"{V_PREFIX}/rules/ipv6",
        headers={"x-access-token": jwt_token},
        json={
            "action": 3,
            "next_header": "tcp",
            "source": "2001:718:1C01:1111::",
            "source_mask": 64,
            "source_port": "",
            "expires": "1444913400",
        },
    )
    data = json.loads(req.data)
    assert req.status_code == 201
    assert data["rule"]
    assert data["rule"]["id"] == "1"
    assert data["rule"]["user"] == "jiri.vrany@cesnet.cz"
    assert data["rule"]["expires"] == 1444913400


def test_create_v6rule_with_timestamp(client, db, jwt_token):
    """
    test that creating with valid data returns 201
    """
    req = client.post(
        f"{V_PREFIX}/rules/ipv6",
        headers={"x-access-token": jwt_token},
        json={
            "action": 2,
            "next_header": "udp",
            "source": "2001:718:1C01:1111::",
            "source_mask": 64,
            "source_port": "",
            "expires": "2549952908",
        },
    )
    data = json.loads(req.data)
    assert req.status_code == 201
    assert data["rule"]
    assert data["rule"]["id"] == "2"
    assert data["rule"]["rstate"] == "active rule"
    assert data["rule"]["user"] == "jiri.vrany@cesnet.cz"
    assert data["rule"]["expires"] == 2549953200


def test_update_existing_rtbh_rule_with_timestamp(client, db, jwt_token):
    """
    test that update with different data passes
    """
    req = client.post(
        f"{V_PREFIX}/rules/rtbh",
        headers={"x-access-token": jwt_token},
        json={
            "community": 1,
            "ipv4": "147.230.17.17",
            "ipv4_mask": 32,
            "expires": "1444913400",
        },
    )
    data = json.loads(req.data)
    assert req.status_code == 201
    assert data["rule"]
    assert data["rule"]["id"] == 1
    assert data["rule"]["user"] == "jiri.vrany@cesnet.cz"
    assert data["rule"]["expires"] == 1444913400


def test_create_rtbh_rule_with_timestamp(client, db, jwt_token):
    """
    test that creating with valid data returns 201
    """
    req = client.post(
        f"{V_PREFIX}/rules/rtbh",
        headers={"x-access-token": jwt_token},
        json={
            "community": 1,
            "ipv4": "147.230.17.117",
            "ipv4_mask": 32,
            "expires": "1444913400",
        },
    )
    data = json.loads(req.data)
    assert req.status_code == 201
    assert data["rule"]
    assert data["rule"]["id"] == 2
    assert data["rule"]["user"] == "jiri.vrany@cesnet.cz"
    assert data["rule"]["expires"] == 1444913400


def test_create_v4rule_lmit(client, db, app, jwt_token):
    """
    test that limit checkt for v4 works
    """
    with app.app_context():
        org = db.session.query(Organization).filter_by(id=1).first()
        org.limit_flowspec4 = 2
        db.session.commit()

        # count
        count = db.session.query(Flowspec4).count()
        print("COUNT", count)

    sources = ["147.230.42.17", "147.230.42.118"]
    codes = [201, 403]

    for source, code in zip(sources, codes):
        data = {
            "action": 1,
            "protocol": "tcp",
            "source": source,
            "source_mask": 32,
            "source_port": "",
            "expires": "10/15/2050 14:46",
        }
        req = client.post(
            f"{V_PREFIX}/rules/ipv4",
            headers={"x-access-token": jwt_token},
            json=data,
        )

        assert req.status_code == code


def test_create_v6rule_lmit(client, db, app, jwt_token):
    """
    test that limit check for v6 works
    """
    with app.app_context():
        org = db.session.query(Organization).filter_by(id=1).first()
        org.limit_flowspec6 = 3
        db.session.commit()

    sources = ["2001:718:1C01:1111::", "2001:718:1C01:1112::"]
    codes = [201, 403]

    for source, code in zip(sources, codes):
        data = {
            "action": 1,
            "next_header": "tcp",
            "source": source,
            "source_mask": 64,
            "source_port": "",
            "expires": "10/15/2050 14:46",
        }
        req = client.post(
            f"{V_PREFIX}/rules/ipv6",
            headers={"x-access-token": jwt_token},
            json=data,
        )

        assert req.status_code == code


def test_create_rtbh_lmit(client, db, app, jwt_token):
    """
    test that limit check for v6 works
    """
    with app.app_context():
        org = db.session.query(Organization).filter_by(id=1).first()
        org.limit_rtbh = 1
        db.session.commit()

    sources = ["147.230.17.42", "147.230.17.43"]
    codes = [201, 403]

    for source, code in zip(sources, codes):
        data = {
            "community": 1,
            "ipv4": source,
            "ipv4_mask": 32,
            "expires": "10/25/2050 14:46",
        }
        req = client.post(f"{V_PREFIX}/rules/rtbh", headers={"x-access-token": jwt_token}, json=data)

        assert req.status_code == code


def test_update_existing_v4rule_with_timestamp_limit(client, db, app, jwt_token):
    """
    test that update with different data passes
    """
    with app.app_context():
        # count
        count = db.session.query(Flowspec4).filter_by(org_id=1, rstate_id=1).count()
        print("COUNT in update", count)

        org = db.session.query(Organization).filter_by(id=1).first()
        org.limit_flowspec4 = count
        db.session.commit()

    req = client.post(
        f"{V_PREFIX}/rules/ipv4",
        headers={"x-access-token": jwt_token},
        json={
            "action": 2,
            "protocol": "tcp",
            "source": "147.230.17.17",
            "source_mask": 32,
            "source_port": "",
            "expires": "2552634908",
        },
    )

    assert req.status_code == 403
    data = json.loads(req.data)
    assert data["message"]
    assert data["message"].startswith("Rule limit")


def test_overall_limit(client, db, app, jwt_token):
    """
    test that update with different data passes
    """
    app.config.update({"FLOWSPEC4_MAX_RULES": 5, "FLOWSPEC6_MAX_RULES": 5, "RTBH_MAX_RULES": 5})

    with app.app_context():
        # count

        org = db.session.query(Organization).filter_by(id=1).first()
        org.limit_flowspec4 = 20
        db.session.commit()

    sources = ["147.230.42.1", "147.230.42.2", "147.230.42.3", "147.230.42.4"]
    codes = [201, 201, 201, 403]

    for source, code in zip(sources, codes):
        data = {
            "action": 1,
            "protocol": "tcp",
            "source": source,
            "source_mask": 32,
            "source_port": "",
            "expires": "10/15/2050 14:46",
        }
        req = client.post(
            f"{V_PREFIX}/rules/ipv4",
            headers={"x-access-token": jwt_token},
            json=data,
        )
        print(source)
        assert req.status_code == code

    data = json.loads(req.data)
    assert data["message"]
    assert data["message"].startswith("System limit")



from flowapp.constants import (
    ANNOUNCE,
    WITHDRAW,
    IPV4_DEFMASK,
    IPV6_DEFMASK,
    MAX_PACKET,
    IPV4_PROTOCOL,
    IPV6_NEXT_HEADER,
)
from flowapp.flowspec import translate_sequence as trps
from flask import current_app
from flowapp.models import ASPath
from flowapp import db


def create_ipv4(rule, message_type=ANNOUNCE):
    """
    create ExaBpg text message for IPv4 rule
    @param rule models.Flowspec4
    @return string message
    """
    protocol = ""
    if rule.protocol and rule.protocol != "all":
        protocol = "protocol ={};".format(IPV4_PROTOCOL[rule.protocol])

    flagstring = rule.flags.replace(";", " ") if rule.flags else ""

    flags = "tcp-flags {};".format(flagstring) if rule.flags and rule.protocol == "tcp" else ""

    fragment_string = rule.fragment.replace(";", " ") if rule.fragment else ""
    fragment = "fragment [ {} ];".format(fragment_string) if rule.fragment else ""

    spec = {
        "protocol": protocol,
        "flags": flags,
        "fragment": fragment,
        "mask": IPV4_DEFMASK,
    }

    return create_message(rule, spec, message_type)


def create_ipv6(rule, message_type=ANNOUNCE):
    """
    create ExaBpg text message for IPv6 rule
    @param rule models.Flowspec6
    @return string message
    :param message_type:
    """
    protocol = ""
    if rule.next_header and rule.next_header != "all":
        protocol = "next-header ={};".format(IPV6_NEXT_HEADER[rule.next_header])
    flagstring = rule.flags.replace(";", " ")
    flags = "tcp-flags {};".format(flagstring) if rule.flags and rule.next_header == "tcp" else ""

    spec = {"protocol": protocol, "mask": IPV6_DEFMASK, "flags": flags}

    return create_message(rule, spec, message_type)


def create_rtbh(rule, message_type=ANNOUNCE):
    """
    create RTBH message in ExaBgp text format
    route 10.10.10.1/32 next-hop 192.0.2.1 community 65001:666 no-export
    """
    action = "announce"
    if message_type == WITHDRAW:
        action = "withdraw"

    if rule.ipv4:
        my_4mask = sanitize_mask(rule.ipv4_mask, IPV4_DEFMASK)
        source = "{}".format(rule.ipv4) if rule.ipv4 else ""
        source += "/{}".format(my_4mask) if rule.ipv4 else ""
        nexthop = "192.0.2.1"

    if rule.ipv6:
        my_6mask = sanitize_mask(rule.ipv6_mask, IPV6_DEFMASK)
        source = "{}".format(rule.ipv6) if rule.ipv6 else ""
        source += "/{}".format(my_6mask) if rule.ipv6 else ""
        nexthop = "100::1"

    try:
        if current_app.config["USE_RD"]:
            rd_string = "rd {rd} label {label}".format(
                rd=current_app.config["RD_STRING"], label=current_app.config["RD_LABEL"]
            )
        else:
            rd_string = ""
    except KeyError:
        rd_string = ""

    try:
        if current_app.config["USE_MULTI_NEIGHBOR"] and rule.community.comm:
            if rule.community.comm in current_app.config["MULTI_NEIGHBOR"].keys():
                targets = current_app.config["MULTI_NEIGHBOR"].get(rule.community.comm)
            else:
                targets = current_app.config["MULTI_NEIGHBOR"].get("primary")

            neighbor = prepare_multi_neighbor(targets)
        else:
            neighbor = ""
    except KeyError:
        neighbor = ""

    community_string = "community [{}]".format(rule.community.comm) if rule.community.comm else ""
    large_community_string = "large-community [{}]".format(rule.community.larcomm) if rule.community.larcomm else ""
    extended_community_string = (
        "extended-community [{}]".format(rule.community.extcomm) if rule.community.extcomm else ""
    )

    as_path_string = ""
    if rule.community.as_path:
        match = db.session.query(ASPath).filter(ASPath.prefix == source).first()
        as_path_string = f"as-path [ {match.as_path} ]" if match else ""

    return "{neighbor}{action} route {source} next-hop {nexthop} {as_path} {community} {large_community} {extended_community}{rd_string}".format(
        neighbor=neighbor,
        action=action,
        source=source,
        as_path=as_path_string,
        community=community_string,
        large_community=large_community_string,
        extended_community=extended_community_string,
        nexthop=nexthop,
        rd_string=rd_string,
    )


def create_message(rule, ipv_specific, message_type=ANNOUNCE):
    """
    create text message using format

    tcp-flagy

    flow route { match { source 147.230.17.6/32;protocol =tcp;tcp-flags fin
    syn;destination-port =3128 >=8080&<=8088;} then {rate-limit 10000; } }

    announce flow route source 4.0.0.0/24 destination 127.0.0.0/24 protocol
    [ udp ] source-port [ =53 ] destination-port [ =80 ]
    packet-length [ =777 =1122 ] fragment [ is-fragment dont-fragment ] rate-limit 1024"


    """
    action = "announce"
    if message_type == WITHDRAW:
        action = "withdraw"

    smask = sanitize_mask(rule.source_mask, ipv_specific["mask"])
    source = "source {}".format(rule.source) if rule.source else ""
    source += "/{};".format(smask) if rule.source else ""

    source_port = "source-port {};".format(trps(rule.source_port)) if rule.source_port else ""

    dmask = sanitize_mask(rule.dest_mask, ipv_specific["mask"])
    dest = " destination {}".format(rule.dest) if rule.dest else ""
    dest += "/{};".format(dmask) if rule.dest else ""

    dest_port = "destination-port {};".format(trps(rule.dest_port)) if rule.dest_port else ""

    protocol = ipv_specific.get("protocol", "")
    flags = ipv_specific.get("flags", "")
    fragment = ipv_specific.get("fragment", "")

    packet_len = "packet-length {};".format(trps(rule.packet_len, MAX_PACKET)) if rule.packet_len else ""

    values = [source, source_port, dest, dest_port, protocol, fragment, flags, packet_len]
    match_body = " ".join(v for v in values if v)

    command = "{};".format(rule.action.command)

    try:
        if current_app.config["USE_RD"]:
            rd_string = "route-distinguisher {rd};".format(rd=current_app.config["RD_STRING"])
            rt_string = "extended-community target:{rt};".format(rt=current_app.config["RT_STRING"])
        else:
            rd_string = ""
            rt_string = ""
    except KeyError:
        rd_string = ""
        rt_string = ""

    try:
        if current_app.config["USE_MULTI_NEIGHBOR"]:
            targets = current_app.config["MULTI_NEIGHBOR"].get("primary")
            neighbor = prepare_multi_neighbor(targets)
        else:
            neighbor = ""
    except KeyError:
        neighbor = ""

    message_body = "{neighbor}{action} flow route {{ {rd_string} match {{ {match_body} }} then {{ {command} {rt_string} }} }}".format(
        neighbor=neighbor,
        action=action,
        match_body=match_body,
        command=command,
        rd_string=rd_string,
        rt_string=rt_string,
    )

    return message_body


def sanitize_mask(rule_mask, default_mask=IPV4_DEFMASK):
    """
    Sanitize mask / prefix of rule
    :param default_mask: default mask to return if mask is not in the rule
    :param rule: flowspec rule
    :return: int mask
    """
    if rule_mask is None:
        return default_mask

    if 0 <= rule_mask <= default_mask:
        return rule_mask
    else:
        return default_mask


def prepare_multi_neighbor(targets: list):
    """
    prepare multi neighbor string
    """
    neigbors = [f"neighbor {tgt}" for tgt in targets]
    return ", ".join(neigbors) + " "



import pytest
import flowapp.forms


@pytest.fixture()
def ip_form(app, field_class):
    with app.test_request_context():  # Push the request context
        form = flowapp.forms.IPForm()
        form.source = field_class()
        form.dest = field_class()
        form.source_mask = field_class()
        form.dest_mask = field_class()
        return form


def test_ip_form_created(ip_form):
    assert ip_form.source.data is None
    assert ip_form.source.errors == []


@pytest.mark.parametrize(
    "address, mask, expected",
    [
        ("147.230.23.25", "24", False),
        ("147.230.23.0", "24", True),
        ("0.0.0.0", "0", True),
        ("2001:718:1C01:1111::1111", "64", False),
        ("2001:718:1C01:1111::", "64", True),
    ],
)
def test_ip_form_validate_source_address(ip_form, address, mask, expected):
    ip_form.source.data = address
    ip_form.source_mask.data = mask
    assert ip_form.validate_source_address() == expected


@pytest.mark.parametrize(
    "address, mask, expected",
    [
        ("147.230.23.25", "24", False),
        ("147.230.23.0", "24", True),
        ("0.0.0.0", "0", True),
        ("2001:718:1C01:1111::1111", "64", False),
        ("2001:718:1C01:1111::", "64", True),
    ],
)
def test_ip_form_validate_dest_address(ip_form, address, mask, expected):
    ip_form.dest.data = address
    ip_form.dest_mask.data = mask
    assert ip_form.validate_dest_address() == expected


@pytest.mark.parametrize(
    "address, mask, ranges, expected",
    [
        ("147.230.23.0", "24", ["147.230.0.0/16", "2001:718:1c01::/48"], True),
        ("0.0.0.0", "0", ["147.230.0.0/16", "2001:718:1c01::/48"], False),
        ("195.113.0.0", "16", ["195.113.0.0/18", "195.113.64.0/21"], False),
    ],
)
def test_ip_form_validate_address_mask(ip_form, address, mask, ranges, expected):
    ip_form.net_ranges = ranges
    ip_form.source.data = address
    ip_form.source_mask.data = mask
    assert ip_form.validate_address_ranges() == expected



import subprocess
from datetime import datetime
from dataclasses import dataclass

from flask import (
    Blueprint,
    current_app,
    render_template,
    render_template_string,
    request,
    session,
    make_response,
    abort,
    redirect,
    url_for,
)

from markupsafe import Markup
from flowapp import models, validators, flowspec
from flowapp.auth import auth_required
from flowapp.constants import (
    SORT_ARG,
    ORDER_ARG,
    DEFAULT_ORDER,
    DEFAULT_SORT,
    SEARCH_ARG,
    RULE_ARG,
    TYPE_ARG,
    ORDSRC_ARG,
    COMP_FUNCS,
    DEFAULT_COUNT_MATCH,
)
from flowapp.utils import active_css_rstate, other_rtypes

dashboard = Blueprint("dashboard", __name__, template_folder="templates")

# Pagination constants
PAGE_ARG = "page"
PER_PAGE_DEFAULT = 50  # Default number of items per page
PER_PAGE_OPTIONS = [25, 50, 100, 200]  # Options for items per page
PER_PAGE_ARG = "per_page"


@dashboard.route("/whois/<string:ip_address>", methods=["GET"])
@auth_required
def whois(ip_address):
    result = subprocess.run(["whois", ip_address], stdout=subprocess.PIPE)
    return render_template(
        "pages/dashboard_whois.html",
        result=result.stdout.decode("utf-8"),
        ip_address=ip_address,
    )


@dashboard.route("/<path:rtype>/<path:rstate>/")
@auth_required
def index(rtype=None, rstate="active"):
    """
    dispatcher object for the dashboard
    :param rtype:  ipv4, ipv6, rtbh, whitelist
    :param rstate: active, expired, all
    :return: view from view factory
    """

    # set first key of dashboard config as default rtype
    if not rtype:
        rtype = next(iter(current_app.config["DASHBOARD"].keys()))

    # params sanitization
    if rtype not in current_app.config["DASHBOARD"].keys():
        return abort(404)
    if rstate not in COMP_FUNCS.keys():
        return abort(404)
    if sum(session["user_role_ids"]) == 1:
        rstate = "active"

    # select view factory function based on user role
    if 3 in session["user_role_ids"]:
        view_factory = create_admin_response
    elif 2 in session["user_role_ids"]:
        view_factory = create_user_response
    else:
        view_factory = create_view_response

    # get the macros for the current rule type from config
    macro_file = current_app.config["DASHBOARD"].get(rtype).get("macro_file", "macros.html")
    macro_tbody = current_app.config["DASHBOARD"].get(rtype).get("macro_tbody", "build_ip_tbody")
    macro_thead = current_app.config["DASHBOARD"].get(rtype).get("macro_thead", "build_rules_thead")
    macro_tfoot = current_app.config["DASHBOARD"].get(rtype).get("macro_tfoot", "build_group_buttons_tfoot")

    data_handler_module = current_app.config["DASHBOARD"].get(rtype).get("data_handler", models)
    data_handler_method = current_app.config["DASHBOARD"].get(rtype).get("data_handler_method", "get_ip_rules")

    # Get pagination parameters
    page = request.args.get(PAGE_ARG, 1, type=int)
    per_page = request.args.get(PER_PAGE_ARG, PER_PAGE_DEFAULT, type=int)

    # Validate per_page
    if per_page not in PER_PAGE_OPTIONS:
        per_page = PER_PAGE_DEFAULT

    # get search query, sort order and sort key from request or session
    get_search_query = request.args.get(SEARCH_ARG, session.get(SEARCH_ARG, ""))
    get_sort_key = request.args.get(SORT_ARG, session.get(SORT_ARG, DEFAULT_SORT))
    get_sort_order = request.args.get(ORDER_ARG, session.get(ORDER_ARG, DEFAULT_ORDER))
    get_sort_order_source = request.args.get(ORDSRC_ARG, session.get(ORDSRC_ARG, ""))

    # store current state for redirects
    session[ORDER_ARG] = get_sort_order
    session[SORT_ARG] = get_sort_key
    session[RULE_ARG] = rstate
    session[TYPE_ARG] = rtype
    session[SEARCH_ARG] = get_search_query

    # switch order when click on the same column twice
    try:
        if session[SORT_ARG] == get_sort_key and get_sort_order_source == "link":
            get_sort_order = "desc" if get_sort_order == "asc" else "asc"
            session[ORDER_ARG] = get_sort_order
    except KeyError:
        get_sort_order = DEFAULT_ORDER

    # get the handler and the data
    handler = getattr(data_handler_module, data_handler_method)

    # Determine if we're searching
    is_searching = bool(get_search_query)

    # Always fetch ALL rules first (no pagination at DB level when searching)
    if is_searching:
        # Get all rules for search
        rules = handler(rtype, rstate, get_sort_key, get_sort_order, paginate=False)

        # Perform search on all rules
        rules = filter_rules(rules, get_search_query)

        # Now paginate the search results in memory
        pagination = paginate_list(rules, page, per_page)
        # Get the slice of rules for current page
        start_idx = (page - 1) * per_page
        end_idx = start_idx + per_page
        rules = rules[start_idx:end_idx]

        # Get counts for other rule types
        count_match = current_app.config["COUNT_MATCH"]
        count_match[rtype] = pagination.total
        for other_rtype in other_rtypes(rtype):
            other_rules = handler(other_rtype, rstate, get_sort_key, get_sort_order, paginate=False)
            other_rules = filter_rules(other_rules, get_search_query)
            count_match[other_rtype] = len(other_rules)
    else:
        # No search - use normal pagination or fetch all
        use_pagination = rstate in ["expired", "all"]

        if use_pagination:
            # Use paginated version from DB
            rules_data = handler(
                rtype, rstate, get_sort_key, get_sort_order, page=page, per_page=per_page, paginate=True
            )
            if isinstance(rules_data, tuple):
                rules, pagination = rules_data
            else:
                rules = rules_data
                pagination = None
        else:
            # Fetch all rules for 'active' state
            rules = handler(rtype, rstate, get_sort_key, get_sort_order, paginate=False)
            pagination = None

        count_match = ""

    # Enrich rules with whitelist information
    rules, whitelist_rule_ids = enrich_rules_with_whitelist_info(rules, rtype)

    allowed_communities = current_app.config["ALLOWED_COMMUNITIES"]

    return view_factory(
        rtype=rtype,
        rstate=rstate,
        rules=rules,
        sort_key=get_sort_key,
        sort_order=get_sort_order,
        table_colspan=current_app.config["DASHBOARD"][rtype]["table_colspan"],
        table_columns=current_app.config["DASHBOARD"][rtype]["table_columns"],
        table_title=f'{current_app.config["DASHBOARD"][rtype]["name"]} rules',
        search_query=get_search_query,
        count_match=count_match,
        macro_file=macro_file,
        macro_tbody=macro_tbody,
        macro_thead=macro_thead,
        macro_tfoot=macro_tfoot,
        whitelist_rule_ids=whitelist_rule_ids,
        allowed_communities=allowed_communities,
        pagination=pagination,
        per_page=per_page,
        per_page_options=PER_PAGE_OPTIONS,
    )


# Add this route to your dashboard.py Blueprint


@dashboard.route("/clear-search")
@auth_required
def clear_search():
    """
    Clear the search query from session and redirect back to the current view.
    """
    # Get current rtype and rstate before clearing
    rtype = session.get(TYPE_ARG, next(iter(current_app.config["DASHBOARD"].keys())))
    rstate = session.get(RULE_ARG, "active")
    sort_key = session.get(SORT_ARG, DEFAULT_SORT)
    sort_order = session.get(ORDER_ARG, DEFAULT_ORDER)

    # Clear the search query from session
    session[SEARCH_ARG] = ""

    # Redirect back to dashboard with current settings but no search
    return redirect(url_for("dashboard.index", rtype=rtype, rstate=rstate, sort=sort_key, order=sort_order))


# Helper functions


@dataclass
class Pagination:
    page: int
    per_page: int
    total: int
    pages: int
    has_prev: bool
    has_next: bool
    prev_num: int = None
    next_num: int = None
    first: int = 0
    last: int = 0


def paginate_list(items, page, per_page):
    """
    Create a pagination object from a list of items.
    This mimics SQLAlchemy's pagination for in-memory lists.

    :param items: List of items to paginate
    :param page: Current page number (1-indexed)
    :param per_page: Number of items per page
    :return: Pagination-like object
    """
    total = len(items)
    pages = (total + per_page - 1) // per_page  # Ceiling division

    has_prev = page > 1
    has_next = page < pages

    prev_num = page - 1 if has_prev else None
    next_num = page + 1 if has_next else None

    # Calculate first and last item numbers for display
    first = (page - 1) * per_page + 1 if total > 0 else 0
    last = min(page * per_page, total)

    pagination = Pagination(
        page=page,
        per_page=per_page,
        total=total,
        pages=pages,
        has_prev=has_prev,
        has_next=has_next,
        prev_num=prev_num,
        next_num=next_num,
        first=first,
        last=last,
    )

    return pagination


def create_dashboard_table_body(
    rules,
    rtype,
    editable=True,
    group_op=True,
    macro_file="macros.html",
    macro_name="build_ip_tbody",
    whitelist_rule_ids=None,
    allowed_communities=None,
):
    """
    create the table body for the dashboard using a jinja2 macro
    :param rules:  list of rules
    :param rtype:  ipv4, ipv6, rtbh
    :param editable: whether rules can be edited
    :param group_op: whether group operations are allowed
    :param macro_file:  the file where the macro is defined
    :param macro_name:  the name of the macro
    :param whitelist_rule_ids: set of rule IDs that were created by a whitelist
    """
    tstring = "{% "
    tstring = tstring + f"from '{macro_file}' import {macro_name}"
    tstring = tstring + " %} {{"
    tstring = (
        tstring + f" {macro_name}(rules, today, editable, group_op, whitelist_rule_ids, allowed_communities) " + "}}"
    )

    dashboard_table_body = render_template_string(
        tstring,
        rules=rules,
        today=datetime.now(),
        editable=editable,
        group_op=group_op,
        whitelist_rule_ids=whitelist_rule_ids or set(),
        allowed_communities=allowed_communities or [],
    )
    return dashboard_table_body


def create_dashboard_table_head(
    rules_columns,
    rtype,
    rstate,
    sort_key,
    sort_order,
    search_query="",
    group_op=True,
    macro_file="macros.html",
    macro_name="build_rules_thead",
):
    """
    create the table head for the dashboard using a jinja2 macro
    :param rules_columns:  list of columns
    :param rtype:  ipv4, ipv6, rtbh
    :param rstate:  active, inactive
    :param sort_key:  the column to sort by
    :param sort_order:  asc or desc
    :param search_query:  the search query
    :param group_op:  group operations allowed
    :param macro_file:  the file where the macro is defined
    :param macro_name:  the name of the macro
    """
    tstring = "{% "
    tstring = tstring + f"from '{macro_file}' import {macro_name}"
    tstring = tstring + " %} {{"
    tstring = (
        tstring + f" {macro_name}(rules_columns, rtype, rstate, sort_key, sort_order, search_query, group_op) " + "}}"
    )

    dashboard_table_head = render_template_string(
        tstring,
        rules_columns=rules_columns,
        rtype=rtype,
        rstate=rstate,
        sort_key=sort_key,
        sort_order=sort_order,
        search_query=search_query,
        group_op=group_op,
    )

    return dashboard_table_head


def create_dashboard_table_foot(colspan=10, macro_file="macros.html", macro_name="build_group_buttons_tfoot"):
    """
    create the table foot for the dashboard using a jinja2 macro
    :param colspan:  the number of columns
    :param macro_file:  the file where the macro is defined
    :param macro_name:  the name of the macro
    """
    tstring = "{% "
    tstring = tstring + f"from '{macro_file}' import {macro_name}"
    tstring = tstring + " %} {{"
    tstring = tstring + f" {macro_name}(colspan) " + "}}"

    dashboard_table_foot = render_template_string(tstring, colspan=colspan)
    return dashboard_table_foot


def create_admin_response(
    rtype,
    rstate,
    rules,
    sort_key,
    sort_order,
    table_colspan,
    table_columns,
    table_title,
    search_query="",
    count_match=DEFAULT_COUNT_MATCH,
    macro_file="macros.html",
    macro_tbody="build_ip_tbody",
    macro_thead="build_rules_thead",
    macro_tfoot="build_group_buttons_tfoot",
    whitelist_rule_ids=None,
    allowed_communities=None,
    pagination=None,
    per_page=PER_PAGE_DEFAULT,
    per_page_options=PER_PAGE_OPTIONS,
):
    """
    Admin can see and edit any rules
    :param pagination: SQLAlchemy pagination object (optional)
    :param per_page: Number of items per page
    :param per_page_options: Available options for items per page
    """
    group_op = True if rtype != "whitelist" else False

    dashboard_table_body = create_dashboard_table_body(
        rules,
        rtype,
        macro_file=macro_file,
        macro_name=macro_tbody,
        group_op=group_op,
        whitelist_rule_ids=whitelist_rule_ids,
        allowed_communities=allowed_communities,
    )

    dashboard_table_head = create_dashboard_table_head(
        rules_columns=table_columns,
        rtype=rtype,
        rstate=rstate,
        sort_key=sort_key,
        sort_order=sort_order,
        search_query=search_query,
        group_op=group_op,
        macro_file=macro_file,
        macro_name=macro_thead,
    )
    if group_op:
        dashboard_table_foot = create_dashboard_table_foot(
            table_colspan,
            macro_file=macro_file,
            macro_name=macro_tfoot,
        )
    else:
        dashboard_table_foot = ""

    res = make_response(
        render_template(
            "pages/dashboard_admin.html",
            display_rules=len(rules),
            table_title=table_title,
            css_classes=active_css_rstate(rtype, rstate),
            count_match=count_match,
            dashboard_table_body=Markup(dashboard_table_body),
            dashboard_table_head=Markup(dashboard_table_head),
            dashboard_table_foot=Markup(dashboard_table_foot),
            rules_columns=table_columns,
            rtype=rtype,
            rstate=rstate,
            sort_key=sort_key,
            sort_order=sort_order,
            search_query=search_query,
            pagination=pagination,
            per_page=per_page,
            per_page_options=per_page_options,
        )
    )

    return res


def create_user_response(
    rtype,
    rstate,
    rules,
    sort_key,
    sort_order,
    table_colspan,
    table_columns,
    table_title,
    search_query="",
    count_match=DEFAULT_COUNT_MATCH,
    macro_file="macros.html",
    macro_tbody="build_ip_tbody",
    macro_thead="build_rules_thead",
    macro_tfoot="build_rules_tfoot",
    whitelist_rule_ids=None,
    allowed_communities=None,
    pagination=None,
    per_page=PER_PAGE_DEFAULT,
    per_page_options=PER_PAGE_OPTIONS,
):
    """
    Filter out the rules for normal users
    :param pagination: SQLAlchemy pagination object (optional)
    :param per_page: Number of items per page
    :param per_page_options: Available options for items per page
    """

    net_ranges = models.get_user_nets(session["user_id"])

    if rtype == "rtbh":
        rules_editable, read_only_rules = validators.split_rtbh_rules_for_user(net_ranges, rules)
    else:
        user_rules, read_only_rules = validators.split_rules_for_user(net_ranges, rules)
        user_actions = models.get_user_actions(session["user_role_ids"])
        user_actions = [act[0] for act in user_actions]
        rules_editable, rules_visible = flowspec.filter_rules_action(user_actions, user_rules)
        read_only_rules = read_only_rules + rules_visible

    # we don't want the read only rules if they are not active
    if rstate != "active":
        read_only_rules = []

    dashboard_table_readonly = create_dashboard_table_body(
        read_only_rules,
        rtype,
        editable=False,
        group_op=False,
        macro_file=macro_file,
        macro_name=macro_tbody,
        whitelist_rule_ids=whitelist_rule_ids,
        allowed_communities=allowed_communities,
    )

    group_op = True if rtype != "whitelist" else False

    dashboard_table_editable = create_dashboard_table_body(
        rules_editable,
        rtype,
        macro_file=macro_file,
        macro_name=macro_tbody,
        group_op=group_op,
        whitelist_rule_ids=whitelist_rule_ids,
        allowed_communities=allowed_communities,
    )
    dashboard_table_editable_head = create_dashboard_table_head(
        rules_columns=table_columns,
        rtype=rtype,
        rstate=rstate,
        sort_key=sort_key,
        sort_order=sort_order,
        search_query=search_query,
        group_op=group_op,
        macro_file=macro_file,
        macro_name=macro_thead,
    )
    dashboard_table_readonly_head = create_dashboard_table_head(
        rules_columns=table_columns,
        rtype=rtype,
        rstate=rstate,
        sort_key=sort_key,
        sort_order=sort_order,
        search_query=search_query,
        group_op=False,
        macro_file=macro_file,
        macro_name=macro_thead,
    )

    if group_op:
        dashboard_table_foot = create_dashboard_table_foot(
            table_colspan,
            macro_file=macro_file,
            macro_name=macro_tfoot,
        )
    else:
        dashboard_table_foot = ""

    display_editable = len(rules_editable)
    display_readonly = len(read_only_rules)

    res = make_response(
        render_template(
            "pages/dashboard_user.html",
            table_title=table_title,
            rules_columns=table_columns,
            dashboard_table_editable=Markup(dashboard_table_editable),
            dashboard_table_readonly=Markup(dashboard_table_readonly),
            display_editable=display_editable,
            display_readonly=display_readonly,
            css_classes=active_css_rstate(rtype, rstate),
            dashboard_table_editable_head=Markup(dashboard_table_editable_head),
            dashboard_table_readonly_head=Markup(dashboard_table_readonly_head),
            dashboard_table_foot=Markup(dashboard_table_foot),
            rtype=rtype,
            rstate=rstate,
            sort_key=sort_key,
            sort_order=sort_order,
            search_query=search_query,
            count_match=count_match,
            pagination=pagination,
            per_page=per_page,
            per_page_options=per_page_options,
        )
    )

    return res


def create_view_response(
    rtype,
    rstate,
    rules,
    sort_key,
    sort_order,
    table_colspan,
    table_columns,
    table_title,
    search_query="",
    count_match=DEFAULT_COUNT_MATCH,
    macro_file="macros.html",
    macro_tbody="build_ip_tbody",
    macro_thead="build_rules_thead",
    macro_tfoot="build_rules_tfoot",
    whitelist_rule_ids=None,
    allowed_communities=None,
    pagination=None,
    per_page=PER_PAGE_DEFAULT,
    per_page_options=PER_PAGE_OPTIONS,
):
    """
    Filter out the rules for normal users
    :param pagination: SQLAlchemy pagination object (optional)
    :param per_page: Number of items per page
    :param per_page_options: Available options for items per page
    """
    dashboard_table_body = create_dashboard_table_body(
        rules,
        rtype,
        editable=False,
        group_op=False,
        macro_file=macro_file,
        macro_name=macro_tbody,
        whitelist_rule_ids=whitelist_rule_ids,
        allowed_communities=allowed_communities,
    )

    dashboard_table_head = create_dashboard_table_head(
        table_columns,
        rtype,
        rstate,
        sort_key,
        sort_order,
        search_query,
        group_op=False,
        macro_file=macro_file,
        macro_name=macro_thead,
    )

    dashboard_table_foot = create_dashboard_table_foot(
        table_colspan,
        macro_file=macro_file,
        macro_name=macro_tfoot,
    )

    res = make_response(
        render_template(
            "pages/dashboard_view.html",
            table_title=table_title,
            rules_columns=table_columns,
            display_rules=len(rules),
            css_classes=active_css_rstate(rtype, rstate),
            search_query=search_query,
            count_match=count_match,
            rstate=rstate,
            rtype=rtype,
            dashboard_table_body=Markup(dashboard_table_body),
            dashboard_table_head=Markup(dashboard_table_head),
            dashboard_table_foot=Markup(dashboard_table_foot),
            pagination=pagination,
            per_page=per_page,
            per_page_options=per_page_options,
        )
    )

    return res


def filter_rules(rules, get_search_query):
    """
    Filter rules based on search query.
    Performs full-text search across all rule fields.

    :param rules: List of rule objects
    :param get_search_query: Search string
    :return: Filtered list of rules
    """
    if not get_search_query:
        return rules

    rules_serialized = [rule.dict() for rule in rules]
    result = []
    search_lower = get_search_query.lower()

    for idx, rule in enumerate(rules_serialized):
        # Create a full text string from all values
        full_text = " ".join(str(val) for val in rule.values())
        if search_lower in full_text.lower():
            result.append(rules[idx])

    return result


def enrich_rules_with_whitelist_info(rules, rule_type):
    """
    Enrich rules with whitelist information from RuleWhitelistCache.

    Args:
        rules: List of rule objects (Flowspec4, Flowspec6, RTBH)
        rule_type: String identifier of rule type ("ipv4", "ipv6", "rtbh")

    Returns:
        Tuple of (rules, whitelist_rule_ids) where whitelist_rule_ids is a set of
        rule IDs that were created by a whitelist.
    """
    from flowapp.models.rules.whitelist import RuleWhitelistCache
    from flowapp.constants import RuleTypes, RuleOrigin

    # Map rule type string to enum value
    rule_type_map = {"ipv4": RuleTypes.IPv4.value, "ipv6": RuleTypes.IPv6.value, "rtbh": RuleTypes.RTBH.value}

    # Get all rule IDs
    rule_ids = [rule.id for rule in rules]

    # No rules to process
    if not rule_ids:
        return rules, set()

    # Query the cache for these rule IDs
    cache_entries = RuleWhitelistCache.query.filter(
        RuleWhitelistCache.rid.in_(rule_ids), RuleWhitelistCache.rtype == rule_type_map.get(rule_type)
    ).all()

    # Create a set of rule IDs that were created by a whitelist
    whitelist_rule_ids = {entry.rid for entry in cache_entries if entry.rorigin == RuleOrigin.WHITELIST.value}

    return rules, whitelist_rule_ids
